/*
 * This file is part of the Molecfit Pipeline
 * Copyright (C) 2001-2019 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
    #include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 *                              Includes
 */
/*----------------------------------------------------------------------------*/

#include "molecfit_config.h"

/*----------------------------------------------------------------------------*/
/**
 *                              Defines
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 *                              Functions prototypes
 */
/*----------------------------------------------------------------------------*/


/* Split a string into pieces at a given delimiter */
static char ** molecfit_config_str_split(
    const char               *str,            /* Input string                               */
    const char               *delimiter,      /* Delimiter to split the string              */
    cpl_size                 *size);          /* Out: Number of tokens                      */

/* Deallocate a string array */
static void molecfit_config_str_array_delete(
    char                     **strarr);       /* Input string array                         */

/*----------------------------------------------------------------------------*/
/**
 *                          Static variables
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup molecfit_config  Common functions to the molecfit recipes
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 *                              Functions code
 */
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
 * @brief    Add parameter to the recipe parameterlist
 *
 * @param    recipe        Name of the recipe
 * @param    plist         Propertylist in the recipe for put the parameter
 * @param    param         Name of the parameter
 * @param    range         Flag: If the value is unique or it's a range
 * @param    rMin          In case of range: Min value
 * @param    rMax          In case of range: Max value
 * @param    type          Type of the parameter
 * @param    strlst_type   Type of string list (only relevant if type is CPL_STRING)
 * @param    value         Generic pointer to the value/s of the parameter
 * @param    description   Description of the parameter
 * @param    mf_conf       Flag: If the parameter is for the molecfit configuration
 *
 * @return   CPL_ERROR_NONE if everything is OK and CPL_ERROR_INVALID_TYPE if
 *           the type isn't correct.
 *
 * Description:
 *    Create a parameter with the input and insert in the recipe parameterlist.
 */
/*----------------------------------------------------------------------------*/
cpl_error_code molecfit_config_fill_parameter(
    const char               *recipe,
    cpl_parameterlist        *plist,
    const char               *param,
    cpl_boolean              range,
    const void               *rMin,
    const void               *rMax,
    cpl_type                 type,
    enum MF_STRLST_TYPE      strlst_type,
    const void               *value,
    const char               *description,
    cpl_boolean              mf_conf)
{
  cpl_parameter *p;
  
  /* Parameter is either:                                           */
  /*                      a single value                            */
  /*                      a "range"                                 */
  /*                      or a string representing a list of values */

  if (range) {

      /* ------------------------ */
      /* PARAMETER TYPE IS RANGE: */
      /* ------------------------ */

      if(type == CPL_TYPE_INT) {

          const int val    = *(const int *)value;
          const int valMin = *(const int *)rMin;
          const int valMax = *(const int *)rMax;
          if(val < valMin || val > valMax) {
              return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal int value range");
          }
          p = cpl_parameter_new_range(param, type, description, recipe, val, valMin, valMax);

      } else if( type == CPL_TYPE_DOUBLE) {

          const double val     = *(const double *)value;
          const double valMin  = *(const double *)rMin;
          const double valMax  = *(const double *)rMax;
          if(val < valMin || val > valMax) {
              return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal double value range");
          }
          p = cpl_parameter_new_range(param, type, description, recipe, val, valMin, valMax);

      } else {
      
          return cpl_error_set_message(cpl_func, CPL_ERROR_INVALID_TYPE, "cpl_type is not correct with ranges");
      }

  } else if (type==CPL_TYPE_STRING && strlst_type!=MF_STRLST_NULL) {

      /* ---------------------------------------- */
      /* PARAMETER TYPE IS STRING LIST OF VALUES: */
      /* ---------------------------------------- */

      const char * strlst = (const char *) value;
      cpl_boolean is_valid = mf_strlst_is_valid(strlst,strlst_type);
      if (!is_valid) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_INVALID_TYPE, "Invalid string list %s",strlst);
      }      
      p = cpl_parameter_new_value(param, type, description, recipe, strlst);
       
  } else {

      /* --------------------------------- */
      /* PARAMETER TYPE IS A SINGLE VALUE: */
      /* --------------------------------- */

      switch(type) {
        case CPL_TYPE_STRING :
          p = cpl_parameter_new_value(param, type, description, recipe,  (const char *)       value); break;
        case CPL_TYPE_BOOL :
          p = cpl_parameter_new_value(param, type, description, recipe, *(const cpl_boolean *)value); break;
        case CPL_TYPE_INT :
          p = cpl_parameter_new_value(param, type, description, recipe, *(const int *)        value); break;
        case CPL_TYPE_DOUBLE :
          p = cpl_parameter_new_value(param, type, description, recipe, *(const double *)     value); break;
        default :
          return cpl_error_set_message(cpl_func, CPL_ERROR_INVALID_TYPE, "cpl_type is not correct");
      }
  }

  if(!mf_conf) {
      cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, param);
      cpl_parameter_disable(  p, CPL_PARAMETER_MODE_ENV);
  } else {
      cpl_parameter_disable(  p, CPL_PARAMETER_MODE_CLI);
      cpl_parameter_disable(  p, CPL_PARAMETER_MODE_ENV);
      cpl_parameter_disable(  p, CPL_PARAMETER_MODE_CFG);
  }
  
  cpl_parameterlist_append(plist, p);

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Add molecfit config parameters to the recipe parameters
 *
 * @param  recipe                        .
 * @param  self                          .
 *
 * @return   cpl_error_code
 */
/*----------------------------------------------------------------------------*/
cpl_error_code molecfit_config_fill_parameterlist(
    const char               *recipe,
    cpl_parameterlist        *self)
{
  /* Add the different default parameters to the recipe */
  cpl_errorstate pre_state = cpl_errorstate_get();
  cpl_error_code e         = CPL_ERROR_NONE;


  /* Fill the parameters list */
  cpl_boolean    range     = CPL_TRUE;
  const void     *dummyMin = NULL;
  const void     *dummyMax = NULL;


  /* --MF_PARAMETERS_TELLURICCORR_PATH */
  if (!e) {
      const char *telluriccorr_path = MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT;
      e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TELLURICCORR_PATH,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)telluriccorr_path,
                                                   MF_PARAMETERS_TELLURICCORR_PATH_DESC, CPL_FALSE);
  }

  /* --MF_PARAMETERS_TELLURICCORR_DATA_PATH */
  if (!e) {
      const char *telluriccorr_data_path = MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT;
      e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TELLURICCORR_DATA_PATH,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)telluriccorr_data_path,
                                                   MF_PARAMETERS_TELLURICCORR_DATA_PATH_DESC, CPL_FALSE);
  }

  /* --MF_PARAMETERS_TMP_PATH */
  if (!e) {
      const char *telluriccorr_tmp_path = MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT;
      e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TMP_PATH,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)telluriccorr_tmp_path,
                                                   MF_PARAMETERS_TMP_PATH_DESC, CPL_FALSE);
  }

  /* --MF_PARAMETERS_OUTPUT_PATH */
  /*if (!e) {
      const char *telluriccorr_output_path = MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT;
      e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OUTPUT_PATH,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, (const void *)telluriccorr_output_path,
                                                   MF_PARAMETERS_OUTPUT_PATH_DESC, CPL_FALSE);
  }*/

  /* --MF_PARAMETERS_OUTPUT_NAME */
  /*if (!e) {
      const char *telluriccorr_output_name = MF_PARAMETERS_OUTPUT_NAME_INIT;
      e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OUTPUT_NAME,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, (const void *)telluriccorr_output_name,
                                                   MF_PARAMETERS_OUTPUT_NAME_DESC, CPL_FALSE);
  }*/



  /* --MF_PARAMETERS_OMP_NUM_THREADS */
  /*int omp_num_threads = 1;   */
  /*if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OMP_NUM_THREADS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, &omp_num_threads,
                                                   MF_PARAMETERS_OMP_NUM_THREADS_DESC" - Non parallel by default", CPL_FALSE);*/

  /* --MF_PARAMETERS_SILENT_EXTERNAL_BINS */
  cpl_boolean silent_external_bins = MF_PARAMETERS_SILENT_EXTERNAL_BINS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_SILENT_EXTERNAL_BINS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &silent_external_bins,
                                                   MF_PARAMETERS_SILENT_EXTERNAL_BINS_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_TRANSMISSION */
  cpl_boolean trans = MF_PARAMETERS_TRANSMISSION_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TRANSMISSION,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &trans,
                                                   MF_PARAMETERS_TRANSMISSION_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_COLUMN_LAMBDA */
  const char *col_lambda = MF_PARAMETERS_COLUMN_LAMBDA_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_COLUMN_LAMBDA,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)col_lambda,
                                                   MF_PARAMETERS_COLUMN_LAMBDA_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_COLUMN_FLUX */
  const char *col_flux = MF_PARAMETERS_COLUMN_FLUX_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_COLUMN_FLUX,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)col_flux,
                                                   MF_PARAMETERS_COLUMN_FLUX_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_COLUMN_DFLUX */
  const char *col_dflux = MF_PARAMETERS_COLUMN_DFLUX_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_COLUMN_DFLUX,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)col_dflux,
                                                   MF_PARAMETERS_COLUMN_DFLUX_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_COLUMN_MASK */
  const char *col_mask = MF_PARAMETERS_COLUMN_MASK_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_COLUMN_MASK,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)col_mask,
                                                   MF_PARAMETERS_COLUMN_MASK_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_DEFAULT_ERROR */
  double default_error = MF_PARAMETERS_DEFAULT_ERROR_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_DEFAULT_ERROR,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &default_error,
                                                   MF_PARAMETERS_DEFAULT_ERROR_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_WLG_TO_MICRON */
  double wlg_to_micron = MF_PARAMETERS_WLG_TO_MICRON_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_WLG_TO_MICRON,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &wlg_to_micron,
                                                   MF_PARAMETERS_WLG_TO_MICRON_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_WAVELENGTH_FRAME */
  const char *wavelengths_frame = MF_PARAMETERS_WAVELENGTH_FRAME_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_WAVELENGTH_FRAME,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)wavelengths_frame,
                                                   MF_PARAMETERS_WAVELENGTH_FRAME_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_OBSERVATORY_ERF_RV_KEYWORD // --MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE */
  const char *observing_erf_rv_keyword = MF_PARAMETERS_OBSERVATORY_ERF_RV_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OBSERVATORY_ERF_RV_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)observing_erf_rv_keyword,
                                                   MF_PARAMETERS_OBSERVATORY_ERF_RV_DESC, CPL_FALSE);
                                                   
  double observing_erf_rv_value = MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &observing_erf_rv_value,
                                                   MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_CLEAN_FLUX */
  cpl_boolean clean_flux = MF_PARAMETERS_CLEAN_MODEL_FLUX_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_CLEAN_MODEL_FLUX,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &clean_flux,
                                                   MF_PARAMETERS_CLEAN_MODEL_FLUX_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_FTOL */
  double ftol = MF_PARAMETERS_FTOL_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_FTOL,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &ftol,
                                                   MF_PARAMETERS_FTOL_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_XTOL */
  double xtol = MF_PARAMETERS_XTOL_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_XTOL,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &xtol,
                                                   MF_PARAMETERS_XTOL_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_FLUX_UNIT */
  int flux_unit = MF_PARAMETERS_FLUX_UNIT_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_FLUX_UNIT,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT,MF_STRLST_NULL,
                                                   &flux_unit,
                                                   MF_PARAMETERS_FLUX_UNIT_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_FIT_TELBACK */
  cpl_boolean fit_telescope_background = MF_PARAMETERS_FIT_TELESCOPE_BACK_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_FIT_TELESCOPE_BACK,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &fit_telescope_background,
                                                   MF_PARAMETERS_FIT_TELESCOPE_BACK_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_TELBACK */
  double telescope_background_const = MF_PARAMETERS_TELESCOPE_BACK_CONST_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TELESCOPE_BACK_CONST,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &telescope_background_const,
                                                   MF_PARAMETERS_TELESCOPE_BACK_CONST_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_FIT_CONTINUUM */
  const char* fit_continuum_str_lst = MF_PARAMETERS_FIT_CONTINUUM_INIT ;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, "FIT_CONTINUUM",
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, MF_STRLST_NULL,
                                                   fit_continuum_str_lst,
                                                   MF_PARAMETERS_FIT_CONTINUUM_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_CONTINUUM_N */
           /* int continuum_n = MF_PARAMETERS_CONTINUUM_N_INIT;*/
  const char* continuum_n_str_lst = MF_PARAMETERS_CONTINUUM_N_INIT ;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, "CONTINUUM_N",
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   continuum_n_str_lst,
                                                   MF_PARAMETERS_CONTINUUM_N_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_CONTINUUM_CONST */
  double continuum_const = MF_PARAMETERS_CONTINUUM_CONST_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_CONTINUUM_CONST,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &continuum_const,
                                                   MF_PARAMETERS_CONTINUUM_CONST_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_MAP_REGIONS_TO_CHIP */
  const char* map_regions_to_chip_str_lst = MF_PARAMETERS_MAP_REGIONS_TO_CHIP_INIT ;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_MAP_REGIONS_TO_CHIP,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   map_regions_to_chip_str_lst,
                                                   MF_PARAMETERS_MAP_REGIONS_TO_CHIP_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_OBSERVATORY_BARY_RV */
  /*double obs_bary_rv = MF_PARAMETERS_OBSERVATORY_BARY_RV_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OBSERVATORY_BARY_RV,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, &obs_bary_rv,
                                                   MF_PARAMETERS_OBSERVATORY_BARY_RV_DESC, CPL_FALSE);
  */

  /* --MF_PARAMETERS_FIT_WLC */
  /*cpl_boolean fit_wlc = MF_PARAMETERS_FIT_WLC_INIT;*/
  const char * fit_wlc_str = MF_PARAMETERS_FIT_WLC_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, "FIT_WLC",
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   fit_wlc_str,
                                                   MF_PARAMETERS_FIT_WLC_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_WLC_N */
  int wlc_n = MF_PARAMETERS_WLC_N_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_WLC_N,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT,MF_STRLST_NULL,
                                                   &wlc_n,
                                                   MF_PARAMETERS_WLC_N_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_WLC_CONST */
  double wlc_const = MF_PARAMETERS_WLC_CONST_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_WLC_CONST,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &wlc_const,
                                                   MF_PARAMETERS_WLC_CONST_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_WLC_REF */
  /*const char *wlc_ref = MF_PARAMETERS_WLC_REF_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_WLC_REF,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, (const void *)wlc_ref,
                                                   MF_PARAMETERS_WLC_REF_DESC, CPL_FALSE); */


  /* --MF_PARAMETERS_FIT_RES_BOX */
  cpl_boolean fit_res_box = MF_PARAMETERS_FIT_RES_BOX_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_FIT_RES_BOX,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &fit_res_box,
                                                   MF_PARAMETERS_FIT_RES_BOX_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_REL_RES_BOX */
  double relres_box_min = MF_PARAMETERS_RES_BOX_MIN;
  double relres_box_max = MF_PARAMETERS_RES_BOX_MAX;
  double relres_box     = MF_PARAMETERS_RES_BOX_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_RES_BOX,
                                                   range, &relres_box_min, &relres_box_max, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &relres_box,
                                                   MF_PARAMETERS_RES_BOX_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_FIT_GAUSS */
  cpl_boolean fit_res_gauss = MF_PARAMETERS_FIT_GAUSS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_FIT_GAUSS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &fit_res_gauss,
                                                   MF_PARAMETERS_FIT_GAUSS_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_RES_GAUSS */
  double res_gauss_min = MF_PARAMETERS_RES_GAUSS_MIN;
  double res_gauss_max = MF_PARAMETERS_RES_GAUSS_MAX;
  double res_gauss     = MF_PARAMETERS_FIT_GAUSS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_RES_GAUSS,
                                                   range, &res_gauss_min, &res_gauss_max, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &res_gauss,
                                                   MF_PARAMETERS_RES_GAUSS_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_RES_LORENTZ */
  cpl_boolean fit_res_lorentz = MF_PARAMETERS_FIT_LORENTZ_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_FIT_LORENTZ,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &fit_res_lorentz,
                                                   MF_PARAMETERS_FIT_LORENTZ_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_FIT_LORENTZ */
  double res_lorentz_min = MF_PARAMETERS_RES_LORENTZ_MIN;
  double res_lorentz_max = MF_PARAMETERS_RES_LORENTZ_MAX;
  double res_lorentz     = MF_PARAMETERS_RES_LORENTZ_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_RES_LORENTZ,
                                                   range, &res_lorentz_min, &res_lorentz_max, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &res_lorentz,
                                                   MF_PARAMETERS_RES_LORENTZ_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_KERN_MODE */
  cpl_boolean kern_mode = MF_PARAMETERS_KERN_MODE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_KERN_MODE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &kern_mode,
                                                   MF_PARAMETERS_KERN_MODE_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_KERN_FAC */
  double kern_fac = MF_PARAMETERS_KERN_FAC_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_KERN_FAC,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &kern_fac,
                                                   MF_PARAMETERS_KERN_FAC_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_VAR_KERN */
  cpl_boolean var_kern = MF_PARAMETERS_VAR_KERN_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_VAR_KERN,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &var_kern,
                                                   MF_PARAMETERS_VAR_KERN_DESC, CPL_FALSE);



  /* --MF_PARAMETERS_OBSERVING_DATE_KEYWORD // --MF_PARAMETERS_OBSERVING_DATE_VALUE */
  const char *observing_date_keyword = MF_PARAMETERS_OBSERVING_DATE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OBSERVING_DATE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)observing_date_keyword,
                                                   MF_PARAMETERS_OBSERVING_DATE_DESC, CPL_FALSE);
  double observing_date_value = MF_PARAMETERS_OBSERVING_DATE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_OBSERVING_DATE_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &observing_date_value,
                                                   MF_PARAMETERS_OBSERVING_DATE_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_UTC_KEYWORD // --MF_PARAMETERS_UTC_VALUE */
  const char *utc_keyword = MF_PARAMETERS_UTC_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_UTC_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)utc_keyword,
                                                   MF_PARAMETERS_UTC_DESC, CPL_FALSE);
  double utc_value = MF_PARAMETERS_UTC_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_UTC_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &utc_value,
                                                   MF_PARAMETERS_UTC_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_TELESCOPE_ANGLE_KEYWORD // --MF_PARAMETERS_TELESCOPE_ANGLE_VALUE */
  const char *telescope_angle_keyword = MF_PARAMETERS_TELESCOPE_ANGLE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TELESCOPE_ANGLE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)telescope_angle_keyword,
                                                   MF_PARAMETERS_TELESCOPE_ANGLE_DESC, CPL_FALSE);
                                                   
  double telescope_angle_min   = MF_PARAMETERS_TELESCOPE_ANGLE_VALUE_MIN;
  double telescope_angle_max   = MF_PARAMETERS_TELESCOPE_ANGLE_VALUE_MAX;
  double telescope_angle_value = MF_PARAMETERS_TELESCOPE_ANGLE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TELESCOPE_ANGLE_VALUE,
                                                   range, &telescope_angle_min, &telescope_angle_max, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &telescope_angle_value,
                                                   MF_PARAMETERS_TELESCOPE_ANGLE_VALUE_DESC, CPL_FALSE);



  /* --MF_PARAMETERS_RELATIVE_HUMIDITY_KEYWORD // --MF_PARAMETERS_RELATIVE_HUMIDITY_VALUE */
  const char *relative_humidity_keyword = MF_PARAMETERS_RELATIVE_HUMIDITY_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_RELATIVE_HUMIDITY_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)relative_humidity_keyword,
                                                   MF_PARAMETERS_RELATIVE_HUMIDITY_DESC, CPL_FALSE);
                                                   
  double relative_humidity_value = MF_PARAMETERS_RELATIVE_HUMIDITY_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_RELATIVE_HUMIDITY_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &relative_humidity_value,
                                                   MF_PARAMETERS_RELATIVE_HUMIDITY_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_PRESSURE_KEYWORD // --MF_PARAMETERS_PRESSURE_VALUE */
  const char *pressure_keyword = MF_PARAMETERS_PRESSURE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_PRESSURE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)pressure_keyword,
                                                   MF_PARAMETERS_PRESSURE_DESC, CPL_FALSE);
                                                   
  double pressure_value = MF_PARAMETERS_PRESSURE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_PRESSURE_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &pressure_value,
                                                   MF_PARAMETERS_PRESSURE_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_TEMPERATURE_KEYWORD // --MF_PARAMETERS_TEMPERATURE_VALUE */
  const char *temperature_keyword = MF_PARAMETERS_TEMPERATURE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TEMPERATURE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)temperature_keyword,
                                                   MF_PARAMETERS_TEMPERATURE_DESC, CPL_FALSE);
  double temperature_value = MF_PARAMETERS_TEMPERATURE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_TEMPERATURE_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &temperature_value,
                                                   MF_PARAMETERS_TEMPERATURE_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_MIRROR_TEMPERATURE_KEYWORD // --MF_PARAMETERS_MIRROR_TEMPERATURE_VALUE */
  const char *mirror_temperature_keyword = MF_PARAMETERS_MIRROR_TEMPERATURE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_MIRROR_TEMPERATURE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)mirror_temperature_keyword,
                                                   MF_PARAMETERS_MIRROR_TEMPERATURE_DESC, CPL_FALSE);
                                                   
  double mirror_temperature_value = MF_PARAMETERS_MIRROR_TEMPERATURE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_MIRROR_TEMPERATURE_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &mirror_temperature_value,
                                                   MF_PARAMETERS_MIRROR_TEMPERATURE_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_ELEVATION_KEYWORD // --MF_PARAMETERS_ELEVATION_VALUE */
  const char *elevation_keyword = MF_PARAMETERS_ELEVATION_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_ELEVATION_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)elevation_keyword,
                                                   MF_PARAMETERS_ELEVATION_DESC, CPL_FALSE);
                                                   
  double elevation_value = MF_PARAMETERS_ELEVATION_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_ELEVATION_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &elevation_value,
                                                   MF_PARAMETERS_ELEVATION_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_LONGITUDE_KEYWORD // --MF_PARAMETERS_LONGITUDE_VALUE */
  const char *longitude_keyword = MF_PARAMETERS_LONGITUDE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_LONGITUDE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)longitude_keyword,
                                                   MF_PARAMETERS_LONGITUDE_DESC, CPL_FALSE);
                                                   
  double longitude_min   = MF_PARAMETERS_LONGITUDE_VALUE_MIN;
  double longitude_max   = MF_PARAMETERS_LONGITUDE_VALUE_MAX;
  double longitude_value = MF_PARAMETERS_LONGITUDE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_LONGITUDE_VALUE,
                                                   range, &longitude_min, &longitude_max, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &longitude_value,
                                                   MF_PARAMETERS_LONGITUDE_VALUE_DESC, CPL_FALSE);



  /* --MF_PARAMETERS_LATITUDE_KEYWORD // --MF_PARAMETERS_LATITUDE_VALUE */
  const char *latitude_keyword = MF_PARAMETERS_LATITUDE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_LATITUDE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)latitude_keyword,
                                                   MF_PARAMETERS_LATITUDE_DESC, CPL_FALSE);
                                                   
  double latitude_min   = MF_PARAMETERS_LATITUDE_VALUE_MIN;
  double latitude_max   = MF_PARAMETERS_LATITUDE_VALUE_MAX;
  double latitude_value = MF_PARAMETERS_LATITUDE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_LATITUDE_VALUE,
                                                   range, &latitude_min, &latitude_max, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &latitude_value,
                                                   MF_PARAMETERS_LATITUDE_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_SLIT_WIDTH_KEYWORD // --MF_PARAMETERS_SLIT_WIDTH_VALUE */
  const char *slit_width_keyword = MF_PARAMETERS_SLIT_WIDTH_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_SLIT_WIDTH_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)slit_width_keyword,
                                                   MF_PARAMETERS_SLIT_WIDTH_DESC, CPL_FALSE);
                                                   
  double slit_width_value = MF_PARAMETERS_SLIT_WIDTH_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_SLIT_WIDTH_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &slit_width_value,
                                                   MF_PARAMETERS_SLIT_WIDTH_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_PIXEL_SCALE_KEYWORD // --MF_PARAMETERS_PIXEL_SCALE_VALUE */
  const char *pixel_scale_keyword = MF_PARAMETERS_PIXEL_SCALE_KEYWORD_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_PIXEL_SCALE_KEYWORD,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)pixel_scale_keyword,
                                                   MF_PARAMETERS_PIXEL_SCALE_DESC, CPL_FALSE);
                                                   
  double pixel_scale_value = MF_PARAMETERS_PIXEL_SCALE_VALUE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_PIXEL_SCALE_VALUE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &pixel_scale_value,
                                                   MF_PARAMETERS_PIXEL_SCALE_VALUE_DESC, CPL_FALSE);


  /* --MF_PARAMETERS_REF_ATM */
  const char *ref_atm = MF_PARAMETERS_REFERENCE_ATMOSPHERIC_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_REFERENCE_ATMOSPHERIC,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)ref_atm,
                                                   MF_PARAMETERS_REFERENCE_ATMOSPHERIC_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_GDAS_PROF */
  const char *gdas_prof = MF_PARAMETERS_GDAS_PROFILE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_GDAS_PROFILE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)gdas_prof,
                                                   MF_PARAMETERS_GDAS_PROFILE_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_LAYERS */
  cpl_boolean layers = MF_PARAMETERS_LAYERS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_LAYERS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL,MF_STRLST_NULL,
                                                   &layers,
                                                   MF_PARAMETERS_LAYERS_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_EMIX */
  double emix = MF_PARAMETERS_EMIX_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_EMIX,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &emix,
                                                   MF_PARAMETERS_EMIX_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_PWV */
  double pwv = MF_PARAMETERS_PWV_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_PWV,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &pwv,
                                                   MF_PARAMETERS_PWV_DESC, CPL_FALSE);



  /* --MF_LNFL_LINE_DB */
  const char *lnfl_line_db = MF_LNFL_LINE_DB_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LNFL_LINE_DB,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)lnfl_line_db,
                                                   MF_LNFL_LINE_DB_DESC, CPL_FALSE);

  /* --MF_LNFL_LINE_DB_FMT */
  double lnfl_line_db_fmt = MF_LNFL_LINE_DB_FMT_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LNFL_LINE_DB_FMT,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &lnfl_line_db_fmt,
                                                   MF_LNFL_LINE_DB_FMT_DESC, CPL_FALSE);



  /* --MF_LBLRTM_ICNTNM */
  int lblrtm_icntnm = MF_LBLRTM_ICNTNM_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_ICNTNM,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT,MF_STRLST_NULL,
                                                   &lblrtm_icntnm,
                                                   MF_LBLRTM_ICNTNM_DESC, CPL_FALSE);

  /* --MF_LBLRTM_IAERSL */
  int lblrtm_iaersl = MF_LBLRTM_IAERSL_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_IAERSL,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT,MF_STRLST_NULL,
                                                   &lblrtm_iaersl,
                                                   MF_LBLRTM_IAERSL_DESC, CPL_FALSE);

  /* --MF_LBLRTM_MPTS */
  int lblrtm_mpts = MF_LBLRTM_MPTS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_MPTS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT,MF_STRLST_NULL,
                                                   &lblrtm_mpts,
                                                   MF_LBLRTM_MPTS_DESC, CPL_FALSE);

  /* --MF_LBLRTM_NPTS */
  int lblrtm_npts = MF_LBLRTM_NPTS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_NPTS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT,MF_STRLST_NULL,
                                                    &lblrtm_npts,
                                                   MF_LBLRTM_NPTS_DESC, CPL_FALSE);

  /* --MF_LBLRTM_V1 */
  double lblrtm_v1 = MF_LBLRTM_V1_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_V1,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &lblrtm_v1,
                                                   MF_LBLRTM_V1_DESC, CPL_FALSE);
  /* --MF_LBLRTM_V2 */
  double lblrtm_v2 = MF_LBLRTM_V2_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_V2,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &lblrtm_v2,
                                                   MF_LBLRTM_V2_DESC, CPL_FALSE);

  /* --MF_LBLRTM_SAMPLE */
  int lblrtm_sample = MF_LBLRTM_SAMPLE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_SAMPLE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT,MF_STRLST_NULL,
                                                   &lblrtm_sample,
                                                   MF_LBLRTM_SAMPLE_DESC, CPL_FALSE);

  /* --MF_LBLRTM_ALFAL0 */
  double lblrtm_alfal0 = MF_LBLRTM_ALFAL0_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_ALFAL0,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &lblrtm_alfal0,
                                                   MF_LBLRTM_ALFAL0_DESC, CPL_FALSE);

  /* --MF_LBLRTM_AVMASS */
  double lblrtm_avmass = MF_LBLRTM_AVMASS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_AVMASS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE,MF_STRLST_NULL,
                                                   &lblrtm_avmass,
                                                   MF_LBLRTM_AVMASS_DESC, CPL_FALSE);

  /* --MF_LBLRTM_DPTMIN */
  double lblrtm_dptmin = MF_LBLRTM_DPTMIN_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_DPTMIN,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_dptmin,
                                                   MF_LBLRTM_DPTMIN_DESC, CPL_FALSE);

  /* --MF_LBLRTM_DPTFAC */
  double lblrtm_dptfac = MF_LBLRTM_DPTFAC_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_DPTFAC,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_dptfac,
                                                   MF_LBLRTM_DPTFAC_DESC, CPL_FALSE);

  /* --MF_LBLRTM_TBOUND */
  double lblrtm_tbound = MF_LBLRTM_TBOUND_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_TBOUND,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_tbound,
                                                   MF_LBLRTM_TBOUND_DESC, CPL_FALSE);

  /* --MF_LBLRTM_SREMIS1 */
  double lblrtm_sremis1 = MF_LBLRTM_SREMIS1_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_SREMIS1,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_sremis1,
                                                   MF_LBLRTM_SREMIS1_DESC, CPL_FALSE);
  /* --MF_LBLRTM_SREMIS2 */
  double lblrtm_sremis2 = MF_LBLRTM_SREMIS2_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_SREMIS2,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_sremis2,
                                                   MF_LBLRTM_SREMIS2_DESC, CPL_FALSE);
  /* --MF_LBLRTM_SREMIS3 */
  double lblrtm_sremis3 = MF_LBLRTM_SREMIS3_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_SREMIS3,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_sremis3,
                                                   MF_LBLRTM_SREMIS3_DESC, CPL_FALSE);

  /* --MF_LBLRTM_SRREFL1 */
  double lblrtm_srrefl1 = MF_LBLRTM_SRREFL1_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_SRREFL1,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_srrefl1,
                                                   MF_LBLRTM_SRREFL1_DESC, CPL_FALSE);
  /* --MF_LBLRTM_SRREFL2 */
  double lblrtm_srrefl2 = MF_LBLRTM_SRREFL2_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_SRREFL2,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_srrefl2,
                                                   MF_LBLRTM_SRREFL2_DESC, CPL_FALSE);
  /* --MF_LBLRTM_SRREFL3 */
  double lblrtm_srrefl3 = MF_LBLRTM_SRREFL3_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_SRREFL3,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_srrefl3,
                                                   MF_LBLRTM_SRREFL3_DESC, CPL_FALSE);

  /* --MF_LBLRTM_MODEL */
  int lblrtm_model = MF_LBLRTM_MODEL_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_MODEL,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
                                                   &lblrtm_model,
                                                   MF_LBLRTM_MODEL_DESC, CPL_FALSE);


  /* --MF_LBLRTM_ITYPE */
  int lblrtm_itype = MF_LBLRTM_ITYPE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_ITYPE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
                                                   &lblrtm_itype,
                                                   MF_LBLRTM_ITYPE_DESC, CPL_FALSE);

  /* --MF_LBLRTM_NOZERO */
  int lblrtm_nozero = MF_LBLRTM_NOZERO_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_NOZERO,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
                                                   &lblrtm_nozero,
                                                   MF_LBLRTM_NOZERO_DESC, CPL_FALSE);

  /* --MF_LBLRTM_NOPRNT */
  int lblrtm_noprnt = MF_LBLRTM_NOPRNT_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_NOPRNT,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
                                                   &lblrtm_noprnt,
                                                   MF_LBLRTM_NOPRNT_DESC, CPL_FALSE);

  /* --MF_LBLRTM_IPUNCH */
  int lblrtm_ipunch = MF_LBLRTM_IPUNCH_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_IPUNCH,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
                                                   &lblrtm_ipunch,
                                                   MF_LBLRTM_IPUNCH_DESC, CPL_FALSE);

  /* --MF_LBLRTM_RE */
  double lblrtm_re = MF_LBLRTM_RE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_RE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_re,
                                                   MF_LBLRTM_RE_DESC, CPL_FALSE);

  /* --MF_LBLRTM_HSPACE */
  double lblrtm_hspace = MF_LBLRTM_HSPACE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_HSPACE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_hspace,
                                                   MF_LBLRTM_HSPACE_DESC, CPL_FALSE);

  /* --MF_LBLRTM_H2 */
  double lblrtm_h2 = MF_LBLRTM_H2_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_H2,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_h2,
                                                   MF_LBLRTM_H2_DESC, CPL_FALSE);

  /* --MF_LBLRTM_RANGE */
  double lblrtm_range = MF_LBLRTM_RANGE_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_RANGE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_range,
                                                   MF_LBLRTM_RANGE_DESC, CPL_FALSE);

  /* --MF_LBLRTM_BETA */
  double lblrtm_beta = MF_LBLRTM_BETA_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_BETA,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_beta,
                                                   MF_LBLRTM_BETA_DESC, CPL_FALSE);

  /* --MF_LBLRTM_LEN */
  int lblrtm_len = MF_LBLRTM_LEN_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_LEN,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
                                                   &lblrtm_len,
                                                   MF_LBLRTM_LEN_DESC, CPL_FALSE);

  /* --MF_LBLRTM_HOBS */
  double lblrtm_hobs = MF_LBLRTM_HOBS_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_HOBS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_hobs,
                                                   MF_LBLRTM_HOBS_DESC, CPL_FALSE);

  /* --MF_LBLRTM_AVTRAT */
  double lblrtm_avtrat = MF_LBLRTM_AVTRAT_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_AVTRAT,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_avtrat,
                                                   MF_LBLRTM_AVTRAT_DESC, CPL_FALSE);

  /* --MF_LBLRTM_TDIFF1 */
  double lblrtm_tdiff1 = MF_LBLRTM_TDIFF1_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_TDIFF1,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_tdiff1,
                                                   MF_LBLRTM_TDIFF1_DESC, CPL_FALSE);
  /* --MF_LBLRTM_TDIFF2 */
  double lblrtm_tdiff2 = MF_LBLRTM_TDIFF2_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_TDIFF2,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_tdiff2,
                                                   MF_LBLRTM_TDIFF2_DESC, CPL_FALSE);

  /* --MF_LBLRTM_ALTD1 */
  double lblrtm_altd1 = MF_LBLRTM_ALTD1_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_ALTD1,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_altd1,
                                                   MF_LBLRTM_ALTD1_DESC, CPL_FALSE);
  /* --MF_LBLRTM_ALTD2 */
  double lblrtm_altd2 = MF_LBLRTM_ALTD2_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_ALTD2,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_altd2,
                                                   MF_LBLRTM_ALTD2_DESC, CPL_FALSE);

  /* --MF_LBLRTM_DELV */
  double lblrtm_delv = MF_LBLRTM_DELV_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_LBLRTM_DELV,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_DOUBLE, MF_STRLST_NULL,
                                                   &lblrtm_delv,
                                                   MF_LBLRTM_DELV_DESC, CPL_FALSE);

  /* CHECK AND ADD EXPERT OPTIONS FIT_CHIP[idx] and FIT_RANGE[idx] */

    /* NEW CHIP AND RANGE SPECIFIC PARAMETERS */
/*
  if (!e) e = molecfit_config_fill_parameter(recipe, self, "MAP_REGIONS_TO_CHIP",
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, str_lst,
                                                   "Assigning ranges to specific chips", CPL_FALSE);

  if (!e) e = molecfit_config_fill_parameter(recipe, self, "FIT_CONTINUUM",
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, str_lst,
                                                   "Flags for fitting continuum in specific regions", CPL_FALSE);

  if (!e) e = molecfit_config_fill_parameter(recipe, self, "CONTINUUM_N",
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, str_lst,
                                                   "Polynomial order for continuum model in specific regions", CPL_FALSE);

  cpl_boolean fit_wlc = MF_PARAMETERS_FIT_WLC_INIT;
  if (!e) e = molecfit_config_fill_parameter(recipe, self, "FIT_WLC",
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL, &fit_wlc,
                                                   MF_PARAMETERS_FIT_WLC_DESC, CPL_FALSE);
*/

   /*--MF_PARAMETERS_ EXPERT MODE FLAGS */
  cpl_boolean expert_mode_opt = MF_PARAMETERS_EXPERT_MODE_INIT; 
  if (!e) e = molecfit_config_fill_parameter(recipe, self, MF_PARAMETERS_EXPERT_MODE,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL, MF_STRLST_NULL,
                                                   &expert_mode_opt,
                                                   MF_PARAMETERS_EXPERT_MODE_DESC, CPL_FALSE);



  /* Check possible errors */
  if (!cpl_errorstate_is_equal(pre_state) || e != CPL_ERROR_NONE) {
      return cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                   "molecfit_config_fill_parameterlist failed!");
  }

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief  Get Molecfit config parameters from the input recipe parameter list
 *
 * @param  list                          .
 * @param  data_primary_header           .
 * @param  pl                            .
 *
 * @return   cpl_error_code
 */
/*----------------------------------------------------------------------------*/
mf_configuration * molecfit_config_get_parameters(
    const cpl_parameterlist  *list,
    const cpl_propertylist   *data_primary_header,
    cpl_propertylist         *pl)
{
  /* Check input */
  cpl_error_ensure(list && data_primary_header && pl, CPL_ERROR_NULL_INPUT,
                   return NULL, "list input is NULL!");

  /* Get preState */
  cpl_errorstate pre_state = cpl_errorstate_get();
  const cpl_parameter *p;


  /*** Molecfit default configuration ***/
  cpl_msg_info(cpl_func, "Load Molecfit default configuration ...");
  mf_configuration *mf_config = mf_configuration_create();
  if (!mf_config) {
      cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_OUTPUT,
                            "Error loading Molecfit default configuration");
      return NULL;
  }


  /*** Set Molecfit configuration with recipe parameters ***/
  mf_parameters_directories  *directories  = &(mf_config->parameters->directories );
  mf_parameters_inputs       *inputs       = &(mf_config->parameters->inputs      );
  mf_parameters_fitting      *fitting      = &(mf_config->parameters->fitting     );
  mf_parameters_ambient      *ambient      = &(mf_config->parameters->ambient     );
  mf_parameters_instrumental *instrumental = &(mf_config->parameters->instrumental);
  mf_parameters_atmospheric  *atmospheric  = &(mf_config->parameters->atmospheric );


  /*** Directories ***/
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TELLURICCORR_PATH);
  if (p) {
      if (strcmp(cpl_parameter_get_string(p), MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT)) {
          if (directories->telluric_path) cpl_free(directories->telluric_path);
          directories->telluric_path = cpl_strdup(cpl_parameter_get_string(p));
          cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELLURICCORR_PATH, directories->telluric_path);
      }
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TELLURICCORR_DATA_PATH);
  if (p) {
      if (strcmp(cpl_parameter_get_string(p), MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT)) {
          if (directories->telluriccorr_data_path) cpl_free(directories->telluriccorr_data_path);
          directories->telluriccorr_data_path = cpl_strdup(cpl_parameter_get_string(p));
          cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELLURICCORR_DATA_PATH, directories->telluriccorr_data_path);
      }
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TMP_PATH);
  if (p) {
      if (strcmp(cpl_parameter_get_string(p), MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT)) {
          if (directories->tmp_path) cpl_free(directories->tmp_path);
          directories->tmp_path = cpl_strdup(cpl_parameter_get_string(p));
          cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TMP_PATH, directories->tmp_path);
      }
  }

  /*p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OUTPUT_PATH);
  if (p) {
      if (strcmp(cpl_parameter_get_string(p), MOLECIFT_TELLURICCORR_PARAMETER_DEFAULT)) {
          if (directories->output_path) cpl_free(directories->output_path);
          directories->output_path = cpl_strdup(cpl_parameter_get_string(p));
          cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OUTPUT_PATH, directories->output_path);
      }
  }*/

  /*p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OUTPUT_NAME);
  if (p) {
      if (directories->output_name) cpl_free(directories->output_name);
      directories->output_name = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OUTPUT_NAME, directories->output_name);
  }*/


  /*** Input parameters ***/

  /* p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OMP_NUM_THREADS);
  if (p) {
      inputs->omp_num_threads = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OMP_NUM_THREADS, inputs->omp_num_threads);
  }*/

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_SILENT_EXTERNAL_BINS);
  if (p) {
      inputs->silent_external_bins = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_SILENT_EXTERNAL_BINS, inputs->silent_external_bins);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TRANSMISSION);
  if (p) {
      inputs->transmission = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TRANSMISSION, inputs->transmission);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_COLUMN_LAMBDA);
  if (p) {
      if (inputs->column_lambda) cpl_free(inputs->column_lambda);
      inputs->column_lambda = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_LAMBDA, inputs->column_lambda);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_COLUMN_FLUX);
  if (p) {
      if (inputs->column_flux) cpl_free(inputs->column_flux);
      inputs->column_flux = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_FLUX, inputs->column_flux);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_COLUMN_DFLUX);
  if (p) {
      if (inputs->column_dflux) cpl_free(inputs->column_dflux);
      inputs->column_dflux = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_DFLUX, inputs->column_dflux);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_COLUMN_MASK);
  if (p) {
      if (inputs->column_mask) cpl_free(inputs->column_mask);
      inputs->column_mask = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_MASK, inputs->column_mask);
  }


  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_DEFAULT_ERROR);
  if (p) {
      inputs->default_error = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_DEFAULT_ERROR, inputs->default_error);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_WLG_TO_MICRON);
  if (p) {
      inputs->wlg_to_micron = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLG_TO_MICRON, inputs->wlg_to_micron);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_WAVELENGTH_FRAME);
  if (p) {
      if (inputs->wavelengths_frame) cpl_free(inputs->wavelengths_frame);
      inputs->wavelengths_frame = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WAVELENGTH_FRAME, inputs->wavelengths_frame);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OBSERVATORY_ERF_RV_KEYWORD);
  if (p) {
      if (inputs->observing_erv_rv.key) cpl_free(inputs->observing_erv_rv.key);
      inputs->observing_erv_rv.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVATORY_ERF_RV_KEYWORD, inputs->observing_erv_rv.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE);
  if (p) {
      inputs->observing_erv_rv.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE, inputs->observing_erv_rv.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_CLEAN_MODEL_FLUX);
  if (p) {
      inputs->clean_flux = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_CLEAN_MODEL_FLUX, inputs->clean_flux);
  }


  /*** Fitting ***/

  /* Convergence criterion */
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_FTOL);
  if (p) {
      fitting->ftol = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FTOL, fitting->ftol);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_XTOL);
  if (p) {
      fitting->xtol = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_XTOL, fitting->xtol);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_FLUX_UNIT);
  if (p) {
      fitting->flux_unit = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FLUX_UNIT, fitting->flux_unit);
  }


  /*** EXPERT FITTING OPTIONS ***/
  /* 
    CHIP and RANGE FIT OPTIONS:
      This if from the old molecfit expert options of:
          fit_chip1 =1 or 0
          fit_chip2 =1 or 0
          fit_chip3 =1 or 0
          fit_chip4 =1 or 0
          fit_range1=1 or 0
          fit_range2=1 or 0
          fit_range3=1 or 0
          fit_range4=1 or 0
      These are currently added for instrument scientist to play with
      and maybe later decide of how to use.
  */
  
  /* FIT_CHIP[1-4]*/
  p = cpl_parameterlist_find_const(list, "FIT_CHIP1");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_CHIP1 = %f. Ignored as this parameter is now obsolete.",dval);
  }

  p = cpl_parameterlist_find_const(list, "FIT_CHIP2");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_CHIP2 = %f. Ignored as this parameter is now obsolete.",dval);
  }

  p = cpl_parameterlist_find_const(list, "FIT_CHIP3");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_CHIP3 = %f. Ignored as this parameter is now obsolete.",dval);
  }

  p = cpl_parameterlist_find_const(list, "FIT_CHIP4");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_CHIP4 = %f. Ignored as this parameter is now obsolete.",dval);
  }
  
  
  /* FIT_RANGE[1-4]*/
  p = cpl_parameterlist_find_const(list, "FIT_RANGE1");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_RANGE1 = %f. Ignored as this parameter is now obsolete.",dval);
  }

  p = cpl_parameterlist_find_const(list, "FIT_RANGE2");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_RANGE2 = %f. Ignored as this parameter is now obsolete.",dval);
  }

  p = cpl_parameterlist_find_const(list, "FIT_RANGE3");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_RANGE3 = %f. Ignored as this parameter is now obsolete.",dval);
  }

  p = cpl_parameterlist_find_const(list, "FIT_RANGE4");
  if (p) {
      double dval = cpl_parameter_get_double(p);
      cpl_msg_info(cpl_func,"FIT_RANGE4 = %f. Ignored as this parameter is now obsolete.",dval);
  }

  /* NOTE: The recipe parameter "FIT_RANGES" is a string of 1 and 0s deliminated with ',' character
           and the fitting->fit_ranges is a boolean vector so that telluriccor doen't have to
           parse the string to a vector of booleans */
  p = cpl_parameterlist_find_const(list, "FIT_CHIPS");
  if (p) {

      /* Get the string as specified in the .rc file */
      const char * flagstr  = cpl_parameter_get_string(p);

      /* Parse the string into a set of 0s and 1s */
      int cnt =0;
      for (int i=0; i<(int)strlen(flagstr) ;i++) {
           if (flagstr[i]!='1' && flagstr[i]!='0' ) continue;
           cpl_msg_info(cpl_func,"CHIP %d -> %c",cnt,flagstr[i]);
           if (flagstr[i]=='1') fitting->fit_chips[cnt]=CPL_TRUE;
           if (flagstr[i]=='0') fitting->fit_chips[cnt]=CPL_FALSE;
           cnt++;
      }
      if (cnt>0) fitting->fit_n_cflags=cnt;
      cpl_msg_info(cpl_func,"N CHIP FLAGS DEFINED = %d",fitting->fit_n_cflags);

      cpl_propertylist_update_string (pl,MF_PARAMETERS_CONTEX_DEFAULT" ""FIT_CHIPS",flagstr);
  }


  /* NOTE: The recipe parameter "FIT_RANGES" is a string of 1 and 0s deliminated with ',' character
           and the fitting->fit_ranges is a boolean vector so that telluriccor doen't have to
           parse the string to a vector of booleans */
  p = cpl_parameterlist_find_const(list, "FIT_RANGES");
  if (p) {


      /* Get the string as specified in the .rc file */
      const char * flagstr  = cpl_parameter_get_string(p);

      /* Parse the string into a set of 0s and 1s */
      int cnt =0;
      for (size_t i=0; i<strlen(flagstr) ;i++) {
           if (flagstr[i]!='1' && flagstr[i]!='0' ) continue;
           cpl_msg_info(cpl_func,"RANGE %d -> %c",cnt,flagstr[i]);
           if (flagstr[i]=='1') fitting->fit_ranges[cnt]=CPL_TRUE;
           if (flagstr[i]=='0') fitting->fit_ranges[cnt]=CPL_FALSE;
           cnt++;
      }
      if (cnt>0) fitting->fit_n_rflags=cnt;

      cpl_propertylist_update_string (pl,MF_PARAMETERS_CONTEX_DEFAULT" ""FIT_RANGES",flagstr);
      
  }



  /* Telelescope background */
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_FIT_TELESCOPE_BACK);
  if (p) {
      fitting->fit_telescope_background.fit = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_TELESCOPE_BACK, fitting->fit_telescope_background.fit);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TELESCOPE_BACK_CONST);
  if (p) {
      fitting->fit_telescope_background.const_val = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELESCOPE_BACK_CONST, fitting->fit_telescope_background.const_val);
  }

 

  /* Telelescope backgroud */
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_FIT_TELESCOPE_BACK);
  if (p) {
      fitting->fit_telescope_background.fit = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_TELESCOPE_BACK, fitting->fit_telescope_background.fit);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TELESCOPE_BACK_CONST);
  if (p) {
      fitting->fit_telescope_background.const_val = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELESCOPE_BACK_CONST, fitting->fit_telescope_background.const_val);
  }



  /* Continuum */
  p = cpl_parameterlist_find_const(list, "FIT_CONTINUUM_OLD");
  if (p) {
      fitting->fit_continuum.fit = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_CONTINUUM, fitting->fit_continuum.fit);
  }

  p = cpl_parameterlist_find_const(list, "CONTINUUM_N_OLD");
  if (p) {
      fitting->fit_continuum.n = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_CONTINUUM_N, fitting->fit_continuum.n);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_CONTINUUM_CONST);
  if (p) {
      fitting->fit_continuum.const_val = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_CONTINUUM_CONST, fitting->fit_continuum.const_val);
  }

  /*p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OBSERVATORY_BARY_RV);
  if (p) {
      fitting->obs_bary_rv = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVATORY_BARY_RV, fitting->obs_bary_rv);
  }
  */


  /* Wavelength solution fit/adjustment */
  p = cpl_parameterlist_find_const(list, "FIT_WLC_OLD");
  if (p) {
      fitting->fit_wavelenght.fit = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_WLC, fitting->fit_wavelenght.fit);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_WLC_N);
  if (p) {
      fitting->fit_wavelenght.n = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLC_N, fitting->fit_wavelenght.n);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_WLC_CONST);
  if (p) {
      fitting->fit_wavelenght.const_val = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLC_CONST, fitting->fit_wavelenght.const_val);
  }

  /* p = cpl_parameterlist_find_const(list, MF_PARAMETERS_WLC_REF);
  if (p) {
      if (fitting->wlc_ref) cpl_free(fitting->wlc_ref);
      fitting->wlc_ref = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLC_REF, fitting->wlc_ref);
  }*/


  /* Default kernel: Boxcar kernel */
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_FIT_RES_BOX);
  if (p) {
      fitting->fit_res_box.fit = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_RES_BOX, fitting->fit_res_box.fit);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_RES_BOX);
  if (p) {
      fitting->fit_res_box.const_val = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RES_BOX, fitting->fit_res_box.const_val);
  }


  /* Default kernel: Gaussian kernel */
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_FIT_GAUSS);
  if (p) {
      fitting->fit_gauss.fit = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_GAUSS, fitting->fit_gauss.fit);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_RES_GAUSS);
  if (p) {
      fitting->fit_gauss.const_val = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RES_GAUSS, fitting->fit_gauss.const_val);
  }


  /* Default kernel: Lorentz kernel */
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_FIT_LORENTZ);
  if (p) {
      fitting->fit_lorentz.fit = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_LORENTZ, fitting->fit_lorentz.fit);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_RES_LORENTZ);
  if (p) {
      fitting->fit_lorentz.const_val = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RES_LORENTZ, fitting->fit_lorentz.const_val);
  }


  /* Default kernels: Generic parameters */
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_KERN_MODE);
  if (p) {
      fitting->kern_mode = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_KERN_MODE, fitting->kern_mode);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_KERN_FAC);
  if (p) {
      fitting->kern_fac = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_KERN_FAC, fitting->kern_fac);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_VAR_KERN);
  if (p) {
      fitting->var_kern = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_VAR_KERN, fitting->var_kern);
  }


  /*** Ambient ***/
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OBSERVING_DATE_KEYWORD);
  if (p) {
      if (ambient->observing_date.key) cpl_free(ambient->observing_date.key);
      ambient->observing_date.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVING_DATE_KEYWORD, ambient->observing_date.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OBSERVING_DATE_VALUE);
  if (p) {
      ambient->observing_date.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVING_DATE_VALUE, ambient->observing_date.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_UTC_KEYWORD);
  if (p) {
      if (ambient->utc.key) cpl_free(ambient->utc.key);
      ambient->utc.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_UTC_KEYWORD, ambient->utc.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_UTC_VALUE);
  if (p) {
      ambient->utc.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_UTC_VALUE, ambient->utc.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TELESCOPE_ANGLE_KEYWORD);
  if (p) {
      if (ambient->telescope_angle.key) cpl_free(ambient->telescope_angle.key);
      ambient->telescope_angle.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELESCOPE_ANGLE_KEYWORD, ambient->telescope_angle.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TELESCOPE_ANGLE_VALUE);
  if (p) {
      ambient->telescope_angle.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELESCOPE_ANGLE_VALUE, ambient->telescope_angle.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_RELATIVE_HUMIDITY_KEYWORD);
  if (p) {
      if (ambient->relative_humidity.key) cpl_free(ambient->relative_humidity.key);
      ambient->relative_humidity.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RELATIVE_HUMIDITY_KEYWORD, ambient->relative_humidity.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_RELATIVE_HUMIDITY_VALUE);
  if (p) {
      ambient->relative_humidity.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RELATIVE_HUMIDITY_VALUE, ambient->relative_humidity.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_PRESSURE_KEYWORD);
  if (p) {
      if (ambient->pressure.key) cpl_free(ambient->pressure.key);
      ambient->pressure.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PRESSURE_KEYWORD, ambient->pressure.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_PRESSURE_VALUE);
  if (p) {
      ambient->pressure.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PRESSURE_VALUE, ambient->pressure.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TEMPERATURE_KEYWORD);
  if (p) {
      if (ambient->temperature.key) cpl_free(ambient->temperature.key);
      ambient->temperature.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TEMPERATURE_KEYWORD, ambient->temperature.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_TEMPERATURE_VALUE);
  if (p) {
      ambient->temperature.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TEMPERATURE_VALUE, ambient->temperature.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_MIRROR_TEMPERATURE_KEYWORD);
  if (p) {
      if (ambient->mirror_temperature.key) cpl_free(ambient->mirror_temperature.key);
      ambient->mirror_temperature.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_MIRROR_TEMPERATURE_KEYWORD, ambient->mirror_temperature.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_MIRROR_TEMPERATURE_VALUE);
  if (p) {
      ambient->mirror_temperature.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_MIRROR_TEMPERATURE_VALUE, ambient->mirror_temperature.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_ELEVATION_KEYWORD);
  if (p) {
      if (ambient->elevation.key) cpl_free(ambient->elevation.key);
      ambient->elevation.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_ELEVATION_KEYWORD, ambient->elevation.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_ELEVATION_VALUE);
  if (p) {
      ambient->elevation.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_ELEVATION_VALUE, ambient->elevation.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_LONGITUDE_KEYWORD);
  if (p) {
      if (ambient->longitude.key) cpl_free(ambient->longitude.key);
      ambient->longitude.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LONGITUDE_KEYWORD, ambient->longitude.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_LONGITUDE_VALUE);
  if (p) {
      ambient->longitude.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LONGITUDE_VALUE, ambient->longitude.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_LATITUDE_KEYWORD);
  if (p) {
      if (ambient->latitude.key) cpl_free(ambient->latitude.key);
      ambient->latitude.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LATITUDE_KEYWORD, ambient->latitude.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_LATITUDE_VALUE);
  if (p) {
      ambient->latitude.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LATITUDE_VALUE, ambient->latitude.value);
  }


  /*** Instrumental ***/
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_SLIT_WIDTH_KEYWORD);
  if (p) {
      if (instrumental->slit_width.key) cpl_free(instrumental->slit_width.key);
      instrumental->slit_width.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_SLIT_WIDTH_KEYWORD, instrumental->slit_width.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_SLIT_WIDTH_VALUE);
  if (p) {
      instrumental->slit_width.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_SLIT_WIDTH_VALUE, instrumental->slit_width.value);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_PIXEL_SCALE_KEYWORD);
  if (p) {
      if (instrumental->pixel_scale.key) cpl_free(instrumental->pixel_scale.key);
      instrumental->pixel_scale.key = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PIXEL_SCALE_KEYWORD, instrumental->pixel_scale.key);
  }
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_PIXEL_SCALE_VALUE);
  if (p) {
      instrumental->pixel_scale.value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PIXEL_SCALE_VALUE, instrumental->pixel_scale.value);
  }


  /*** Atmospheric ***/
  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_REFERENCE_ATMOSPHERIC);
  if (p) {
      if (atmospheric->ref_atm) cpl_free(atmospheric->ref_atm);
      atmospheric->ref_atm = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_REFERENCE_ATMOSPHERIC, atmospheric->ref_atm);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_GDAS_PROFILE);
  if (p) {
      if (atmospheric->gdas_prof) cpl_free(atmospheric->gdas_prof);
      atmospheric->gdas_prof = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_GDAS_PROFILE, atmospheric->gdas_prof);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_LAYERS);
  if (p) {
      atmospheric->layers = cpl_parameter_get_bool(p);
      cpl_propertylist_update_bool(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LAYERS, atmospheric->layers);
  }


  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_EMIX);
  if (p) {
      atmospheric->emix = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_EMIX, atmospheric->emix);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_PWV);
  if (p) {
      atmospheric->pwv = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PWV, atmospheric->pwv);
  }


  /*** LNFL ***/
  mf_io_lnfl_config *lnfl = mf_config->lnfl;

  p = cpl_parameterlist_find_const(list, MF_LNFL_LINE_DB);
  if (p) {
      if (lnfl->line_db) cpl_free(lnfl->line_db);
      lnfl->line_db = cpl_strdup(cpl_parameter_get_string(p));
      cpl_propertylist_update_string(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LNFL_LINE_DB, lnfl->line_db);
  }

  p = cpl_parameterlist_find_const(list, MF_LNFL_LINE_DB_FMT);
  if (p) {
      lnfl->line_db_fmt = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LNFL_LINE_DB_FMT, lnfl->line_db_fmt);
  }


  /*** LBLRTM ***/
  mf_io_lblrtm_config *lblrtm = mf_config->lblrtm;

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_ICNTNM);
  if (p) {
      lblrtm->icntnm = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ICNTNM, lblrtm->icntnm);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_IAERSL);
  if (p) {
      lblrtm->iaersl = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_IAERSL, lblrtm->iaersl);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_MPTS);
  if (p) {
      lblrtm->mpts = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_MPTS, lblrtm->mpts);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_NPTS);
  if (p) {
      lblrtm->npts = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_NPTS, lblrtm->npts);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_V1);
  if (p) {
      lblrtm->v[0] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_V1, lblrtm->v[0]);
  }
  p = cpl_parameterlist_find_const(list, MF_LBLRTM_V2);
  if (p) {
      lblrtm->v[1] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_V2, lblrtm->v[1]);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_SAMPLE);
  if (p) {
      lblrtm->sample = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SAMPLE, lblrtm->sample);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_ALFAL0);
  if (p) {
      lblrtm->alfal0 = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ALFAL0, lblrtm->alfal0);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_AVMASS);
  if (p) {
      lblrtm->avmass = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_AVMASS, lblrtm->avmass);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_DPTMIN);
  if (p) {
      lblrtm->dptmin = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_DPTMIN, lblrtm->dptmin);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_DPTFAC);
  if (p) {
      lblrtm->dptfac = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_DPTFAC, lblrtm->dptfac);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_TBOUND);
  if (p) {
      lblrtm->tbound = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_TBOUND, lblrtm->tbound);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_SREMIS1);
  if (p) {
      lblrtm->sremis[0] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SREMIS1, lblrtm->sremis[0]);
  }
  p = cpl_parameterlist_find_const(list, MF_LBLRTM_SREMIS2);
  if (p) {
      lblrtm->sremis[1] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SREMIS2, lblrtm->sremis[1]);
  }
  p = cpl_parameterlist_find_const(list, MF_LBLRTM_SREMIS3);
  if (p) {
      lblrtm->sremis[2] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SREMIS3, lblrtm->sremis[2]);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_SRREFL1);
  if (p) {
      lblrtm->srrefl[0] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SRREFL1, lblrtm->srrefl[0]);
  }
  p = cpl_parameterlist_find_const(list, MF_LBLRTM_SRREFL2);
  if (p) {
      lblrtm->srrefl[1] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SRREFL2, lblrtm->srrefl[1]);
  }
  p = cpl_parameterlist_find_const(list, MF_LBLRTM_SRREFL3);
  if (p) {
      lblrtm->srrefl[2] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SRREFL3, lblrtm->srrefl[2]);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_MODEL);
  if (p) {
      lblrtm->model = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_MODEL, lblrtm->model);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_ITYPE);
  if (p) {
      lblrtm->itype = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ITYPE, lblrtm->itype);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_NOZERO);
  if (p) {
      lblrtm->nozero = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_NOZERO, lblrtm->nozero);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_NOPRNT);
  if (p) {
      lblrtm->noprnt = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_NOPRNT, lblrtm->noprnt);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_IPUNCH);
  if (p) {
      lblrtm->ipunch = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_IPUNCH, lblrtm->ipunch);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_RE);
  if (p) {
      lblrtm->re = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_RE, lblrtm->re);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_HSPACE);
  if (p) {
      lblrtm->hspace = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_HSPACE, lblrtm->hspace);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_LATITUDE_VALUE);
  if (p) {
      lblrtm->ref_lat = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LATITUDE_VALUE, lblrtm->ref_lat);
  }

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_ELEVATION_VALUE);
  if (p) {
      lblrtm->h[0] = cpl_parameter_get_double(p)/1000.0 ;
      /*cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_H1, lblrtm->h[0]);*/
  }


  p = cpl_parameterlist_find_const(list, MF_LBLRTM_H2);
  if (p) {
      lblrtm->h[1] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_H2, lblrtm->h[1]);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_RANGE);
  if (p) {
      lblrtm->range = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_RANGE, lblrtm->range);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_BETA);
  if (p) {
      lblrtm->beta = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_BETA, lblrtm->beta);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_LEN);
  if (p) {
      lblrtm->len = cpl_parameter_get_int(p);
      cpl_propertylist_update_int(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_LEN, lblrtm->len);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_HOBS);
  if (p) {
      lblrtm->hobs = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_HOBS, lblrtm->hobs);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_AVTRAT);
  if (p) {
      lblrtm->avtrat = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_AVTRAT, lblrtm->avtrat);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_TDIFF1);
  if (p) {
      lblrtm->tdiff[0] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_TDIFF1, lblrtm->tdiff[0]);
  }
  p = cpl_parameterlist_find_const(list, MF_LBLRTM_TDIFF2);
  if (p) {
      lblrtm->tdiff[1] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_TDIFF2, lblrtm->tdiff[1]);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_ALTD1);
  if (p) {
      lblrtm->altd[0] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ALTD1, lblrtm->altd[0]);
  }
  p = cpl_parameterlist_find_const(list, MF_LBLRTM_ALTD2);
  if (p) {
      lblrtm->altd[1] = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ALTD2, lblrtm->altd[1]);
  }

  p = cpl_parameterlist_find_const(list, MF_LBLRTM_DELV);
  if (p) {
      lblrtm->delv = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_DELV, lblrtm->delv);
  }



  /*** Update the Molecfit configuration with the header ***/
  cpl_error_code e = mf_parameters_config_update_with_header_keywords(mf_config->parameters, data_primary_header);

 

  /* Check possible errors */
  if (!cpl_errorstate_is_equal(pre_state) || e != CPL_ERROR_NONE) {
      mf_configuration_delete(mf_config);
      cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_OUTPUT,
                            "Error loading Molecfit default configuration");
      return NULL;
  }

 return mf_config;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief  delete the memory allocations that are allocated
 *         from calling molecfit_config_load_header_parameters. 
 *         Note: this is primarily used for unit tests
 *
 * @param  mf_configuration     mf_configuration object created from a call
 *                              to molecfit_config_load_header_parameters
 *
 * @return void
 */
/*----------------------------------------------------------------------------*/
void molecfit_config_delete_configuration(mf_configuration* cfg) 
{
    mf_configuration_delete(cfg);
}



/*----------------------------------------------------------------------------*/
/**
 * @brief  Parse output header and add to the mf_parameters_config and the header propertylist
 *
 * @param  parameters_header      Header with input parameters.
 * @param  data_header            Header with the data parameters.
 * @param  pl                     Output header.
 *
 * @return mf_configuration       Molecfit configuration
 */
/*----------------------------------------------------------------------------*/

mf_configuration * molecfit_config_load_header_parameters(
    const cpl_propertylist   *parameters_header,
    const cpl_propertylist   *data_header,
    cpl_propertylist         *pl)
{
  /* Check input */
  cpl_error_ensure(parameters_header && data_header && pl, CPL_ERROR_NULL_INPUT,
                   return NULL, "Input is NULL!");

  /* Get preState */
  cpl_errorstate pre_state = cpl_errorstate_get();
  const char *p;


  /*** Get the molecfit parameters from the BEST_FIT_PARAMETERS primary header (cpl_propertylist *) ***/
  cpl_msg_info(cpl_func, "Loading input parameters saved in a output FITS header ...");

  /*** Molecfit default configuration ***/
  cpl_msg_info(cpl_func, "Load Molecfit default configuration ...");
  mf_configuration *mf_config = mf_configuration_create();
  if (!mf_config) {
      cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_OUTPUT,
                            "Error loading Molecfit default configuration");
      return NULL;
  }

  /*** Set Molecfit configuration with recipe parameters ***/
  mf_parameters_directories  *directories  = &(mf_config->parameters->directories );
  mf_parameters_inputs       *inputs       = &(mf_config->parameters->inputs      );
  mf_parameters_fitting      *fitting      = &(mf_config->parameters->fitting     );
  mf_parameters_ambient      *ambient      = &(mf_config->parameters->ambient     );
  mf_parameters_instrumental *instrumental = &(mf_config->parameters->instrumental);
  mf_parameters_atmospheric  *atmospheric  = &(mf_config->parameters->atmospheric );


  /*** Directories ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELLURICCORR_PATH;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (directories->telluric_path) cpl_free(directories->telluric_path);
      directories->telluric_path = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, directories->telluric_path);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELLURICCORR_DATA_PATH;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (directories->telluriccorr_data_path) cpl_free(directories->telluriccorr_data_path);
      directories->telluriccorr_data_path = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, directories->telluriccorr_data_path);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TMP_PATH;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (directories->tmp_path) cpl_free(directories->tmp_path);
      directories->tmp_path = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, directories->tmp_path);
  }

  /*p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OUTPUT_PATH;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (directories->output_path) cpl_free(directories->output_path);
      directories->output_path = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, directories->output_path);
  }*/

  /*p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OUTPUT_NAME;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (directories->output_name) cpl_free(directories->output_name);
      directories->output_name = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, directories->output_name);
  }*/


  /*** Input parameters ***/
  /*p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OMP_NUM_THREADS;
  if (cpl_propertylist_has(parameters_header, p)) {
      inputs->omp_num_threads = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, inputs->omp_num_threads);
  }*/

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_SILENT_EXTERNAL_BINS;
  if (cpl_propertylist_has(parameters_header, p)) {
      inputs->silent_external_bins = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, inputs->silent_external_bins);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TRANSMISSION;
  if (cpl_propertylist_has(parameters_header, p)) {
      inputs->transmission = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, inputs->transmission);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_LAMBDA;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (inputs->column_lambda) cpl_free(inputs->column_lambda);
      inputs->column_lambda = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, inputs->column_lambda);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_FLUX;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (inputs->column_flux) cpl_free(inputs->column_flux);
      inputs->column_flux = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, inputs->column_flux);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_DFLUX;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (inputs->column_dflux) cpl_free(inputs->column_dflux);
      inputs->column_dflux = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, inputs->column_dflux);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_MASK;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (inputs->column_mask) cpl_free(inputs->column_mask);
      inputs->column_mask = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, inputs->column_mask);
  }


  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_DEFAULT_ERROR;
  if (cpl_propertylist_has(parameters_header, p)) {
      inputs->default_error = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, inputs->default_error);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLG_TO_MICRON;
  if (cpl_propertylist_has(parameters_header, p)) {
      inputs->wlg_to_micron = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, inputs->wlg_to_micron);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WAVELENGTH_FRAME;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (inputs->wavelengths_frame) cpl_free(inputs->wavelengths_frame);
      inputs->wavelengths_frame = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, inputs->wavelengths_frame);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVATORY_ERF_RV_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (inputs->observing_erv_rv.key) cpl_free(inputs->observing_erv_rv.key);
      inputs->observing_erv_rv.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, inputs->observing_erv_rv.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      inputs->observing_erv_rv.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, inputs->observing_erv_rv.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_CLEAN_MODEL_FLUX;
  if (cpl_propertylist_has(parameters_header, p)) {
      inputs->clean_flux = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, inputs->clean_flux);
  }



  /*** Convergence criterion ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FTOL;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->ftol = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->ftol);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_XTOL;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->xtol = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->xtol);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FLUX_UNIT;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->flux_unit = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, fitting->flux_unit);
  }


  /*** Telescope background ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_TELESCOPE_BACK;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_telescope_background.fit = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->fit_telescope_background.fit);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELESCOPE_BACK_CONST;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_telescope_background.const_val = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_int(pl, p, fitting->fit_telescope_background.const_val);
  }


  /*** Continuum ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_CONTINUUM;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_continuum.fit = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->fit_continuum.fit);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_CONTINUUM_N;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_continuum.n = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, fitting->fit_continuum.n);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_CONTINUUM_CONST;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_continuum.const_val = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->fit_continuum.const_val);
  }

  /*p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVATORY_BARY_RV;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->obs_bary_rv = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->obs_bary_rv);
  }
  */


  /*** Wavelength solution fit/adjustment ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_WLC;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_wavelenght.fit = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->fit_wavelenght.fit);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLC_N;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_wavelenght.n = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, fitting->fit_wavelenght.n);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLC_CONST;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_wavelenght.const_val = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->fit_wavelenght.const_val);
  }

  /* p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WLC_REF;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (fitting->wlc_ref) cpl_free(fitting->wlc_ref);
      fitting->wlc_ref = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, fitting->wlc_ref);
  } */


  /*** Default kernel: Boxcar kernel ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_RES_BOX;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_res_box.fit = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->fit_res_box.fit);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RES_BOX;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_res_box.const_val = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->fit_res_box.const_val);
  }


  /*** Default kernel: Gaussian kernel ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_GAUSS;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_gauss.fit = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->fit_gauss.fit);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RES_GAUSS;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_gauss.const_val = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->fit_gauss.const_val);
  }


  /*** Default kernel: Lorentz kernel ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_FIT_LORENTZ;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_lorentz.fit = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->fit_lorentz.fit);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RES_LORENTZ;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->fit_lorentz.const_val = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->fit_lorentz.const_val);
  }


  /*** Default kernels: Generic parameters ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_KERN_MODE;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->kern_mode = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->kern_mode);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_KERN_FAC;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->kern_fac = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, fitting->kern_fac);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_VAR_KERN;
  if (cpl_propertylist_has(parameters_header, p)) {
      fitting->var_kern = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, fitting->var_kern);
  }


  /*** Ambient ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVING_DATE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->observing_date.key) cpl_free(ambient->observing_date.key);
      ambient->observing_date.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->observing_date.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVING_DATE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->observing_date.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->observing_date.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_UTC_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->utc.key) cpl_free(ambient->utc.key);
      ambient->utc.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->utc.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_UTC_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->utc.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->utc.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELESCOPE_ANGLE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->telescope_angle.key) cpl_free(ambient->telescope_angle.key);
      ambient->telescope_angle.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->telescope_angle.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TELESCOPE_ANGLE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->telescope_angle.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->telescope_angle.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RELATIVE_HUMIDITY_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->relative_humidity.key) cpl_free(ambient->relative_humidity.key);
      ambient->relative_humidity.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->relative_humidity.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_RELATIVE_HUMIDITY_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->relative_humidity.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->relative_humidity.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PRESSURE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->pressure.key) cpl_free(ambient->pressure.key);
      ambient->pressure.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->pressure.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PRESSURE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->pressure.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->pressure.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TEMPERATURE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->temperature.key) cpl_free(ambient->temperature.key);
      ambient->temperature.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->temperature.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_TEMPERATURE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->temperature.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->temperature.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_MIRROR_TEMPERATURE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->mirror_temperature.key) cpl_free(ambient->mirror_temperature.key);
      ambient->mirror_temperature.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->mirror_temperature.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_MIRROR_TEMPERATURE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->mirror_temperature.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->mirror_temperature.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_ELEVATION_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->elevation.key) cpl_free(ambient->elevation.key);
      ambient->elevation.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->elevation.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_ELEVATION_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->elevation.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->elevation.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LONGITUDE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->longitude.key) cpl_free(ambient->longitude.key);
      ambient->longitude.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->longitude.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LONGITUDE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->longitude.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->longitude.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LATITUDE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (ambient->latitude.key) cpl_free(ambient->latitude.key);
      ambient->latitude.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, ambient->latitude.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LATITUDE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      ambient->latitude.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, ambient->latitude.value);
  }


  /*** Instrumental ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_SLIT_WIDTH_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (instrumental->slit_width.key) cpl_free(instrumental->slit_width.key);
      instrumental->slit_width.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, instrumental->slit_width.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_SLIT_WIDTH_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      instrumental->slit_width.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, instrumental->slit_width.value);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PIXEL_SCALE_KEYWORD;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (instrumental->pixel_scale.key) cpl_free(instrumental->pixel_scale.key);
      instrumental->pixel_scale.key = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, instrumental->pixel_scale.key);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PIXEL_SCALE_VALUE;
  if (cpl_propertylist_has(parameters_header, p)) {
      instrumental->pixel_scale.value = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, instrumental->pixel_scale.value);
  }


  /*** Atmospheric ***/
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_REFERENCE_ATMOSPHERIC;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (atmospheric->ref_atm) cpl_free(atmospheric->ref_atm);
      atmospheric->ref_atm = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, atmospheric->ref_atm);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_GDAS_PROFILE;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (atmospheric->gdas_prof) cpl_free(atmospheric->gdas_prof);
      atmospheric->gdas_prof = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, atmospheric->gdas_prof);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LAYERS;
  if (cpl_propertylist_has(parameters_header, p)) {
      atmospheric->layers = cpl_propertylist_get_bool(parameters_header, p);
      cpl_propertylist_update_bool(pl, p, atmospheric->layers);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_EMIX;
  if (cpl_propertylist_has(parameters_header, p)) {
      atmospheric->emix = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, atmospheric->emix);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_PWV;
  if (cpl_propertylist_has(parameters_header, p)) {
      atmospheric->pwv = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, atmospheric->pwv);
  }


  /*** LNFL ***/
  mf_io_lnfl_config *lnfl = mf_config->lnfl;

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LNFL_LINE_DB;
  if (cpl_propertylist_has(parameters_header, p)) {
      if (lnfl->line_db) cpl_free(lnfl->line_db);
      lnfl->line_db = cpl_strdup(cpl_propertylist_get_string(parameters_header, p));
      cpl_propertylist_update_string(pl, p, lnfl->line_db);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LNFL_LINE_DB_FMT;
  if (cpl_propertylist_has(parameters_header, p)) {
      lnfl->line_db_fmt = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lnfl->line_db_fmt);
  }


  /*** LBLRTM ***/
  mf_io_lblrtm_config *lblrtm = mf_config->lblrtm;

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ICNTNM;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->icntnm = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->icntnm);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_IAERSL;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->iaersl = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->iaersl);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_MPTS;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->mpts = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->mpts);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_NPTS;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->npts = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->npts);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_V1;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->v[0] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->v[0]);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_V2;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->v[1] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->v[1]);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SAMPLE;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->sample = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->sample);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ALFAL0;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->alfal0 = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->alfal0);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_AVMASS;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->avmass = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->avmass);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_DPTMIN;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->dptmin = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->dptmin);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_DPTFAC;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->dptfac = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->dptfac);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_TBOUND;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->tbound = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->tbound);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SREMIS1;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->sremis[0] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->sremis[0]);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SREMIS2;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->sremis[1] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->sremis[1]);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SREMIS3;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->sremis[2] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->sremis[2]);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SRREFL1;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->srrefl[0] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->srrefl[0]);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SRREFL2;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->srrefl[1] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->srrefl[1]);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_SRREFL3;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->srrefl[2] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->srrefl[2]);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_MODEL;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->model = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->model);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ITYPE;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->itype = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->itype);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_NOZERO;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->nozero = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->nozero);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_NOPRNT;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->noprnt = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->noprnt);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_IPUNCH;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->ipunch = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->ipunch);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_RE;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->re = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->re);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_HSPACE;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->hspace = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->hspace);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_LATITUDE_VALUE; // MF_LBLRTM_REF_LAT;   // MF_PARAMETERS_LATITUDE_VALUE
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->ref_lat = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->ref_lat);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_ELEVATION_VALUE; // MF_LBLRTM_H1;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->h[0] =  ambient->elevation.value /1000.00 ; // cpl_propertylist_get_double(parameters_header, p);
      /*cpl_propertylist_update_double(pl, p, lblrtm->h[0]);*/
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_H2;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->h[1] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->h[1]);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_RANGE;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->range = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->range);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_BETA;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->beta = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->beta);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_LEN;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->len = cpl_propertylist_get_int(parameters_header, p);
      cpl_propertylist_update_int(pl, p, lblrtm->len);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_HOBS;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->hobs = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->hobs);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_AVTRAT;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->avtrat = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->avtrat);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_TDIFF1;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->tdiff[0] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->tdiff[0]);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_TDIFF2;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->tdiff[1] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->tdiff[1]);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ALTD1;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->altd[0] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->altd[0]);
  }
  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_ALTD2;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->altd[1] = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->altd[1]);
  }

  p = MF_PARAMETERS_CONTEX_DEFAULT" "MF_LBLRTM_DELV;
  if (cpl_propertylist_has(parameters_header, p)) {
      lblrtm->delv = cpl_propertylist_get_double(parameters_header, p);
      cpl_propertylist_update_double(pl, p, lblrtm->delv);
  }


  /*** Update the Molecfit configuration with the header ***/
  cpl_error_code e = mf_parameters_config_update_with_header_keywords(mf_config->parameters, data_header);


  cpl_msg_info(cpl_func, "Columns : wave[%s], flux[%s], dflux[%s], mask[%s]", inputs->column_lambda, inputs->column_flux, inputs->column_dflux, inputs->column_mask);


  /* Check possible errors */
  if (!cpl_errorstate_is_equal(pre_state) || e != CPL_ERROR_NONE) {
      mf_configuration_delete(mf_config);
      cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_OUTPUT,
                            "Error loading Molecfit default configuration");
      return NULL;
  }

  return mf_config;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Build the molecules Molecfit cpl_table.
 *
 * @param list_molec_str     .
 * @param fit_molec_str      .
 * @param rel_col_str        .
 *
 * @return cpl_error_code    CPL_ERROR_NONE is everything is OK.
 *                           If not, these are the errors:
 *                           - .
 *                           - Error in subroutine (see subroutines).
 *
 * @note .
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_table * molecfit_config_table_molecules(
    const char               *list_molec_str,
    const char               *fit_molec_str,
    const char               *rel_col_str)
{
  cpl_errorstate prestate = cpl_errorstate_get();
  cpl_error_code err      = CPL_ERROR_NONE;

  cpl_size n_molec;

  char **list_molecs = molecfit_config_str_split(list_molec_str, ",", &n_molec);
  if (!list_molecs || n_molec <= 0) {
      err += cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                                   "There has to be at least one molecule specified in the LIST_MOLEC parameter");
  }
  for (cpl_size i = 0; i < n_molec && !err; i++) {
      if (mf_molecules_str_check(list_molecs[i]) != CPL_ERROR_NONE) {
          err += cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                       "Invalid molecule %s specified in the LIST_MOLEC parameter",list_molecs[i]);
      }
  }

  char **fit_molecs = molecfit_config_str_split(fit_molec_str,  ",", &n_molec);
  if (!fit_molecs || n_molec <= 0) {
      err += cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                                   "There has to be at least one boolean flag specified in the FIT_MOLEC parameter");
  }
  for (cpl_size i = 0; i < n_molec; i++) {
      int fit_molec = atoi(fit_molecs[i]);
      if (fit_molec != 0 && fit_molec != 1) {
          err += cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                                       "Unexpected value %d in the FIT_MOLEC list parameter. Expected a boolean, 0 or 1", fit_molec);
      }
  }

  char **rel_cols = molecfit_config_str_split(rel_col_str,     ",", &n_molec);
  if (!rel_cols || n_molec <= 0) {
      err += cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                                   "It is necessary to have at least one relative value specified in the REL_COL parameter.");
  }
  for (cpl_size i = 0; i < n_molec; i++) {
      double rel_col = strtod(rel_cols[i], NULL);
      if (rel_col < 0.) {
          err += cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                                       "Unexpected value %f in the REL_COL list parameter. It must be >= 0.",rel_col);
      }
  }

  /* Check fails */
  if (err != CPL_ERROR_NONE) {
      if (list_molecs) molecfit_config_str_array_delete(list_molecs);
      if (fit_molecs ) molecfit_config_str_array_delete(fit_molecs );
      if (rel_cols   ) molecfit_config_str_array_delete(rel_cols   );
      return NULL;
  }

  /* Create Molecules */
  cpl_table *molectab = cpl_table_new(n_molec);
  cpl_table_new_column(molectab, MF_COL_LIST_MOLECULES, CPL_TYPE_STRING);
  cpl_table_new_column(molectab, MF_COL_FIT_MOLECULES,  CPL_TYPE_INT   );
  cpl_table_new_column(molectab, MF_COL_REL_COL,        CPL_TYPE_DOUBLE);

  /* Insert molecules properties in the cpl_table */
  for (cpl_size i = 0; i < n_molec; i++) {

      int    fit    = atoi(fit_molecs[i]);
      double relcol = strtod(rel_cols[i], NULL);

      cpl_table_set_string(molectab, MF_COL_LIST_MOLECULES, i, list_molecs[i]);
      cpl_table_set_int(   molectab, MF_COL_FIT_MOLECULES,  i, fit           );
      cpl_table_set_double(molectab, MF_COL_REL_COL,        i, relcol        );

      cpl_msg_debug(cpl_func, "(mf_molecules ) Molecule[%s] -> Fit[%d] : rel_col[%lf]", list_molecs[i], fit, relcol);
  }

  /* Cleanup */
  molecfit_config_str_array_delete(list_molecs);
  molecfit_config_str_array_delete(fit_molecs );
  molecfit_config_str_array_delete(rel_cols   );

  /* Check possible errors */
  if (!cpl_errorstate_is_equal(prestate) || cpl_error_get_code() != CPL_ERROR_NONE || !molectab) {

      cpl_error_set_message(cpl_func, cpl_error_get_code(),
                            "Building cpl_table molecules failed!");
      cpl_table_delete(molectab);
      return NULL;

  } else {

      cpl_msg_info(cpl_func, "cpl_table *molecules loaded :");
      //cpl_table_dump(molectab, 0, cpl_table_get_nrow(molectab), NULL);
  }

  return molectab;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Build the wave/pixel range Molecfit cpl_table
 *
 * @param range_str          .
 * @param type               .
 *
 * @return cpl_error_code    CPL_ERROR_NONE is everything is OK.
 *                           If not, these are the errors:
 *                           - .
 *                           - Error in subroutine (see subroutines).
 *
 * @note .
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_table * molecfit_config_table_ranges(
    const char              *range_str,
    cpl_type                type)
{
  cpl_errorstate prestate = cpl_errorstate_get();

  /*** Wavelength ranges ***/
  cpl_size n_range_tokens;
  char     **ranges = molecfit_config_str_split(range_str, ",", &n_range_tokens);
  cpl_size n_range = n_range_tokens / 2;

  /* Create rangetab */
  cpl_table *rangetab = cpl_table_new(n_range);
  cpl_table_new_column(rangetab, MF_COL_WAVE_RANGE_LOWER, type);
  cpl_table_new_column(rangetab, MF_COL_WAVE_RANGE_UPPER, type);

  for (cpl_size i = 0; i < n_range; i++) {

      cpl_size idx = i * 2;

      if (type == CPL_TYPE_DOUBLE) {

          double start = strtod(ranges[idx],     NULL);
          double end   = strtod(ranges[idx + 1], NULL);

          cpl_table_set_double(rangetab, MF_COL_WAVE_RANGE_LOWER, i, start);
          cpl_table_set_double(rangetab, MF_COL_WAVE_RANGE_UPPER, i, end  );

          cpl_msg_info(cpl_func, "range[%lld] = {%g, %g}", i, start, end);

      } else if (type == CPL_TYPE_INT) {

          cpl_size start = atoi(ranges[idx]    );
          cpl_size end   = atoi(ranges[idx + 1]);

          cpl_table_set_int(rangetab, MF_COL_WAVE_RANGE_LOWER, i, start);
          cpl_table_set_int(rangetab, MF_COL_WAVE_RANGE_UPPER, i, end);

          cpl_msg_info(cpl_func, "range[%lld] = {%lld, %lld}", i, start, end);

      } else {

          cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                "Building cpl_table ranges failed!");
          cpl_table_delete(rangetab);
          return NULL;
      }
  }
  

  /* Cleanup */
  molecfit_config_str_array_delete(ranges);
  cpl_msg_info(cpl_func, "Done deleting");
  

  /* Check possible errors */
  if (!cpl_errorstate_is_equal(prestate) || cpl_error_get_code() != CPL_ERROR_NONE) {

      cpl_error_set_message(cpl_func, cpl_error_get_code(),
                            "Building cpl_table ranges failed!");
      cpl_table_delete(rangetab);
      return NULL;

  } else if (rangetab) {

      cpl_msg_info(cpl_func, "cpl_table *rangetab loaded :");
      //cpl_table_dump(rangetab, 0, cpl_table_get_nrow(rangetab), NULL);
  }

  return rangetab;
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief Build one mapping cpl_table
 *
 * @param mapping_str        .
 * @param column_name        .
 *
 * @return cpl_error_code    CPL_ERROR_NONE is everything is OK.
 *                           If not, these are the errors:
 *                           - .
 *                           - Error in subroutine (see subroutines).
 *
 * @note .
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_table * molecfit_config_table_mapping(
    const char              *mapping_str,
    const char              *column_name)
{
  cpl_errorstate prestate = cpl_errorstate_get();

  cpl_size n_tokens;
  char     **tokens = molecfit_config_str_split(mapping_str, ",", &n_tokens);

  /* Create mappingtab */
  cpl_table *mappingtab = cpl_table_new(n_tokens);
  cpl_table_new_column(mappingtab, column_name, CPL_TYPE_INT);

  for (cpl_size i = 0; i < n_tokens; i++) {
      cpl_size ext_num = atoi(tokens[i]);
      cpl_table_set_int(mappingtab, column_name, i, ext_num);
      cpl_msg_info(cpl_func, "Column[%s] : Mapping input extension[%lld] to {%lld}", column_name, i, ext_num);
  }

  /* Cleanup */
  molecfit_config_str_array_delete(tokens);

  /* Check possible errors */
  if (!cpl_errorstate_is_equal(prestate) || cpl_error_get_code() != CPL_ERROR_NONE) {
      cpl_error_set_message(cpl_func, cpl_error_get_code(),
                            "Building cpl_table mapping failed!");
      cpl_table_delete(mappingtab);
      return NULL;

  } else if (mappingtab) {

      cpl_msg_info(cpl_func, "cpl_table *mappingtab loaded :");
      //cpl_table_dump(mappingtab, 0, cpl_table_get_nrow(mappingtab), NULL);
  }

  return mappingtab;
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given string list has valid syntax for its type
 *
 * @param str         The string to split.
 * @param delimiter   String specifying the locations where to split.
 * @param size        (Output) If non NULL, the number of found tokens is returned.
 *
 * @return err 
 *
 *
 */
/* ---------------------------------------------------------------------------*/

cpl_error_code  molecfit_config_strlstchk(
    const cpl_parameterlist* parlist,
    const char* parname,
    molecfit_strlst_type strlst_type)

{
    cpl_error_code err;
    const cpl_parameter* par;

    err=CPL_ERROR_NONE;
    
 
    /* Check that the parameter name is legal */
    if (!err) {
        par=cpl_parameterlist_find_const(parlist,parname);    
        if (!par) {
            err=CPL_ERROR_ILLEGAL_INPUT;
            cpl_msg_error(cpl_func,"Invalid parameter name %s",parname);
            return err;
        }
    }

    /* Check that the string list is a valid list of the specified element type */
    const char* strlst=NULL;
    if (!err) {
    
        strlst=cpl_parameter_get_string(cpl_parameterlist_find_const(parlist,parname));

        if (mf_strlst_is_null(strlst)) {
            cpl_msg_info(cpl_func,"Syntax format for parameter %s OK (=NULL)", parname);
            return err;
        }

        
        switch(strlst_type) {
            case(SLIST):
               if ( ! mf_strlst_is_valid (strlst,MF_STRLST_STRING) ) {
                   cpl_msg_error(cpl_func,"Bad string list specified for parameter %s",parname);
                   return CPL_ERROR_ILLEGAL_INPUT;
               }
               break;
               
            case(DLIST):
            case(DRANGE):
               if ( ! mf_strlst_is_valid (strlst,MF_STRLST_DOUBLE) ) {
                   cpl_msg_error(cpl_func,"Bad double list specified for parameter %s",parname);
                   return CPL_ERROR_ILLEGAL_INPUT;
               }
               break;
               
            case(NLIST):
            case(N0LIST):
            case(NRANGE):
            case(N0RANGE):
               if ( ! mf_strlst_is_valid (strlst,MF_STRLST_INT) ) {
                   cpl_msg_error(cpl_func,"Bad integer list specified for parameter %s",parname);
                   return CPL_ERROR_ILLEGAL_INPUT;
               }
               break;
               
            case(BLIST_OLD):
                if (strlst[0]=='T' || strlst[0]=='t' || strlst[0]=='F' || strlst[0]=='f' ) {
                    break;
                } else if ( ! mf_strlst_is_valid (strlst,MF_STRLST_BOOL) ) {
                   cpl_msg_error(cpl_func,"Bad boolean list specified for parameter %s",parname);
                   return CPL_ERROR_ILLEGAL_INPUT;
                }
		       break;
	       	
            case(BLIST):
               if ( ! mf_strlst_is_valid (strlst,MF_STRLST_BOOL) ) {
                   cpl_msg_error(cpl_func,"Bad boolean list specified for parameter %s",parname);
                   return CPL_ERROR_ILLEGAL_INPUT;
               }
               break;
                              
        };/*end switch strlst_type*/
    }
    
    if (err!=CPL_ERROR_NONE) return err;

    /* If the type is a range list then check that there is an even number of elements */
    int nelems =  mf_strlst_get_size(strlst);
    if (strlst_type==DRANGE||strlst_type==NRANGE||strlst_type==N0RANGE) {
        if (nelems%2!=0) {
            cpl_msg_error(cpl_func,"Range list parameter %s has odd number of elements",parname);
            err=CPL_ERROR_ILLEGAL_INPUT;
            return err;
        }
    }
 
    
    /* If N list or N range check that all values are positive */
    if (strlst_type==NRANGE || strlst_type==NLIST) {  
        cpl_boolean  non_positive_integer=CPL_FALSE;
        for (int idx=1; idx<=nelems; idx++) {
           int ival =  mf_strlst_get_int(strlst,idx);
           if (ival<1) non_positive_integer=CPL_TRUE;
        }
        if (non_positive_integer) {
            cpl_msg_error(cpl_func,"Range list parameter %s has invalid non positive values",parname);
            err=CPL_ERROR_ILLEGAL_INPUT;
            return err;
        }
    }
        
        
    /* If N0 list or N0 range check that all values are positive or zero*/
    if (strlst_type==N0RANGE || strlst_type==N0LIST)  { 
        cpl_boolean  negative_integer=CPL_FALSE;
        for (int idx=1; idx<=nelems; idx++) {
           int ival =  mf_strlst_get_int(strlst,idx);
           if (ival<0) negative_integer=CPL_TRUE;
        }
        if (negative_integer) {
            cpl_msg_error(cpl_func,"Range list parameter %s has invalid negative values",parname);
            err=CPL_ERROR_ILLEGAL_INPUT;
            return err;
        }
    }
    cpl_msg_info(cpl_func,"Syntax format for parameter %s  OK", parname);
    return err;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given string is one of a list of string values
 *
 * @param str         The string to split.
 * @param delimiter   String specifying the locations where to split.
 * @param size        (Output) If non NULL, the number of found tokens is returned.
 *
 * @return err 
 *
 *
 */
/* ---------------------------------------------------------------------------*/

cpl_error_code  molecfit_config_stroptchk(
    const cpl_parameterlist* parlist,
    const char* parname,
    const char* options_strlst)
{    
    cpl_error_code err;
    const cpl_parameter* par;

    err=CPL_ERROR_NONE;
    
    
    /* Check that the parameter name is legal */
    if (!err) {
        par=cpl_parameterlist_find_const(parlist,parname);
        if (!par) {
            cpl_msg_error(cpl_func,"Invalid parameter %s.",parname);
            return CPL_ERROR_ILLEGAL_INPUT;
        }
    }
    
    /* Check that the options string list is a valid string list */
    if (!err) {
        if (! mf_strlst_is_valid (options_strlst,MF_STRLST_STRING) ) {
            cpl_msg_error(cpl_func,"Invalid string list %s",options_strlst);
            return CPL_ERROR_ILLEGAL_INPUT;
        }
    }
        
    
    /* Now check that the parameter string value is in the string list*/    
    if (!err) {
    
        /* Get the parameter as a string value */
        const char* par_str=cpl_parameter_get_string(cpl_parameterlist_find_const(parlist,parname));
        int parstr_size=strlen(par_str);
    
        /* Iterate through the options string list to see if the list contains this string value*/
        int nelems = mf_strlst_get_size(options_strlst);
        cpl_boolean found=CPL_FALSE;
        char str_elem[parstr_size];
        for (int idx=1;idx<=nelems;idx++) {
            cpl_boolean flag = mf_strlst_get_str(options_strlst,idx,str_elem);
            if (flag==CPL_FALSE) continue;
            int str_size  = strlen(str_elem);
            if (str_size!=parstr_size) continue;
            
            /* Strings are same length so check if each character is identical*/
            cpl_boolean is_same = CPL_TRUE;
            for (int i=0;i<parstr_size;i++) {
                if (str_elem[i]!=par_str[i]) is_same=CPL_FALSE;
            }
            if (is_same) {
                found=CPL_TRUE;
                break;
            }
        }
        
        /* If not found then falg illegal input error */
        if (!found) {
          cpl_msg_error(cpl_func,"Invalid option %s. Parameter value of %s must be one of: %s",par_str,parname,options_strlst);
          err=CPL_ERROR_ILLEGAL_INPUT;
        }
    }

    if ( err) cpl_msg_error(cpl_func,"%s parameter is not ok",parname);
      
    cpl_msg_info(cpl_func,"%s parameter is  ok",parname);

    return err;
}

cpl_boolean  molecfit_config_str_isin_strlst(
    const char* inp_str,
    const char* options_strlst)
{    
     /* Iterate through the options string list to see if the list contains this string value*/
     int inp_str_size  = strlen(inp_str);
     int nelems = mf_strlst_get_size(options_strlst);
     cpl_boolean found=CPL_FALSE;
     int options_strlst_size=strlen(options_strlst);
     char str_elem[options_strlst_size];
     for (int idx=1;idx<=nelems;idx++) {
         cpl_boolean flag = mf_strlst_get_str(options_strlst,idx,str_elem);
         if (flag==CPL_FALSE) continue;
         int str_size  = strlen(str_elem);
         if (str_size!= inp_str_size) continue;

         /* Strings are the same length so check that they are identical */
         cpl_boolean is_same = CPL_TRUE;
         for (int i=0;i<inp_str_size;i++) {
             if (str_elem[i]!=inp_str[i]) is_same=CPL_FALSE;
         }
         
	     if (is_same) {
             found=CPL_TRUE;
             break;
         }
     }

    return found;

}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Checks if given string list is actually "NULL" (case insensitive)
 *           Return 1 if true and 0 if false.
 *
 * @param  strlst                 The string list to check 
 
*/
/*----------------------------------------------------------------------------*/

cpl_boolean  mf_strlst_is_null     (const char* strlst) {

/*
    int const true  = 1;
    int const false = 0;
*/
    /* Check for trivial case that string is empty*/
    if (strlst[0]=='\0') return CPL_FALSE;
    
    
    int string_size = strlen(strlst);
    
    /* Check if string is all blanks and if not where first non blank is*/
    cpl_boolean all_blanks=CPL_TRUE;
    int idx0=-1;
    for (int i=0;i<string_size;i++) {
        if (strlst[i]==' ') continue;
        all_blanks=CPL_FALSE;
        idx0=i;
        break;       
    }
    
    /* If string is all blanks then return false */    
    if (all_blanks) return CPL_FALSE;
    
    
    /* Check that there are enough characters in string to spell null*/
    if (idx0+3>string_size) return CPL_FALSE;
    
    /* Check that this and the next 3 characters spell NULL independant of case*/
    if (strlst[idx0+0]!='N' && strlst[idx0+0]!='n') return CPL_FALSE; 
    if (strlst[idx0+1]!='U' && strlst[idx0+1]!='u') return CPL_FALSE; 
    if (strlst[idx0+2]!='L' && strlst[idx0+2]!='l') return CPL_FALSE; 
    if (strlst[idx0+3]!='L' && strlst[idx0+3]!='l') return CPL_FALSE; 
    
    /* We dont care what comes after the word NULL so return true*/
    return CPL_TRUE;

}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Returns the mumber of elements in the given string list
 *
 * @param  strlst                 The string list to check 
 
*/
/*----------------------------------------------------------------------------*/


int  mf_strlst_get_size    (const char* strlst) {

    /* Check for trivial case that string is empty*/
    if (strlst[0]=='\0') return 0;

    /* String is non empty so count the number of deliminators*/
    int size = strlen(strlst);
    int cnt=0;
    for (int i=0;i<size;i++) {
        if (strlst[i]==',') cnt=cnt+1;
    }
    
    /* No of list elements is the no of deliminators +1*/
    return cnt+1;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Checks if given string list is valid given a specific string list type
 *           Return 1 if true and 0 if false.
 *
 * @param  strlst                 The string list to check 
 * @param  strlst_type            The string list type. 
 
*/
/*----------------------------------------------------------------------------*/

cpl_boolean mf_strlst_is_valid (const char* strlst, enum MF_STRLST_TYPE strlst_type) {

    cpl_boolean valid_flag;

    switch (strlst_type) {
    
        case(MF_STRLST_STRING):

            valid_flag=mf_strlst_is_valid_list_of_types(strlst,CPL_TYPE_STRING);
            break;

        case(MF_STRLST_DOUBLE):

            valid_flag=mf_strlst_is_valid_list_of_types(strlst,CPL_TYPE_DOUBLE);
            break;

        case(MF_STRLST_INT):

            valid_flag=mf_strlst_is_valid_list_of_types(strlst,CPL_TYPE_INT);
            break;

        case(MF_STRLST_BOOL):

            valid_flag=mf_strlst_is_valid_list_of_types(strlst,CPL_TYPE_BOOL);
            break;
            
        case(MF_STRLST_DRANGE):
            
            /* Check that string is a list is of valid doubles */
            valid_flag=mf_strlst_is_valid_list_of_types(strlst,CPL_TYPE_DOUBLE);
            
            /* Check that list has an even number of elements */
            if (valid_flag) {
            
                int nelems=mf_strlst_get_size(strlst);
                if (nelems%2 != 0) {
                    cpl_msg_info(cpl_func,"Range list has odd number of elements: %s", strlst);
                    valid_flag=CPL_FALSE;                    
                } 
            }

            /* Check that range pairs [val,val2] is such that val2>val1 */
            if (valid_flag) {
                int nelems=mf_strlst_get_size(strlst);
                for (int i=0;i<=nelems/2;i++) {
                    double val1 = mf_strlst_get_double(strlst,i  );
                    double val2 = mf_strlst_get_double(strlst,i+1);
                    if (val2<val1) {
                        cpl_msg_info(cpl_func,"Range list contains invalid range: [%f,%f]",val1,val2);
                        valid_flag=CPL_FALSE;                    
                    }
                }
            }            
            
            break;
        
        default:
        
            valid_flag = CPL_TRUE;
        
    };/* end switch */
    
    return valid_flag;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Checks if given string list is valid given a specific data type
 *           Return 1 if true and 0 if false.
 *
 * @param  strlst                 The string list to check 
 * @param  strlst                 The data type that the list is made of. 
 
*/
/*----------------------------------------------------------------------------*/

cpl_boolean mf_strlst_is_valid_list_of_types (const char* strlst, cpl_type data_type) {


    /* Check for trivial case that string is empty*/
    if (strlst[0]=='\0') {
        cpl_msg_info(cpl_func,"Invalid empty string list");
        return CPL_FALSE;
    }

    /* Check if NULL */
    if (mf_strlst_is_null(strlst)) return CPL_TRUE;
    
    
    /* Get the no of elements that the string contains*/
    int nelems   = mf_strlst_get_size(strlst);
    /* if no elements then return CPL_FALSE*/
    if (nelems==0)  {
        cpl_msg_info(cpl_func,"Invalid empty string list");
        return CPL_FALSE;
    }
    
    /* Check that each element is non empty */
    int strlst_size=strlen(strlst);
    char elem_str[strlst_size];
    for (int elem=1; elem<=nelems; elem++) {
        cpl_boolean flag=mf_strlst_get_str(strlst,elem,elem_str);
        int elem_size=strlen(elem_str);
	    if (elem_size==0 || flag==CPL_FALSE) {
            cpl_msg_info(cpl_func,"Invalid list member (empty) %d in list string %s ",elem , strlst);
        }
    
    }


    /* If data type is a string then there are no more tests*/
    if (data_type==CPL_TYPE_STRING) return CPL_TRUE;
    
    
    /* Check that each element is of correct type */

    cpl_boolean is_valid=CPL_TRUE; /* Assume all is valid until proven otherwise*/

    for (int elem=1; elem<=nelems; elem++) {

        /* Get the element as a string */
        mf_strlst_get_str(strlst,elem,elem_str);


        /* Attempt a string to double conversion using strtod()*/
        char * p;
        double dval0=strtod (elem_str, &p); /* If succesful conversion then *p=='\0]*/
        if (*p != '\0') {
        
            /* This is a non numerical string */
            cpl_msg_info(cpl_func,"Non numerical element: \"%s\" found as item %d in list: \"%s\"",elem_str,elem,strlst);
            is_valid=CPL_FALSE;
        
        } else {
        

            /* truncate and convert to integer the parsed double dval0*/
            double tdval = trunc(dval0);
            int    ival  = (int) tdval;

            /* Check that this number is valid to the specified data type */
            switch (data_type) {

                case CPL_TYPE_INT:
                if (tdval != dval0) {
                    cpl_msg_info(cpl_func,"Non integer element: \"%s\" found as item %d in list: \"%s\"",elem_str,elem,strlst);
                    is_valid=CPL_FALSE;
                }
                break;

                case CPL_TYPE_BOOL:
                if ( ( ival!=1 && ival!=0 ) || tdval != dval0 ) {
                    cpl_msg_info(cpl_func,"Non boolean element: \"%s\" found as item %d in list: \"%s\"",elem_str,elem,strlst);
                    is_valid=CPL_FALSE;
                }
                break;
                
                default:
                break;

             }
         
         }         
         
    };/* end elem loop*/

     
    return is_valid;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief    Returns the requested element of a string list as a double
 *           
 *
 * @param  strlst                 The string list
 * @param  idx                    The element index 
 
*/
/*----------------------------------------------------------------------------*/

double mf_strlst_get_double (const char* strlst, int idx) {

    /* Get the list element as a string */
    int strlst_size=strlen(strlst);
    char elem_str[strlst_size];
    cpl_boolean flag = mf_strlst_get_str (strlst,idx,elem_str);
    if (flag==CPL_FALSE) return 0.0;

    /* Convert string to a double */
    char *p;
    double dval = strtod (elem_str, &p);
  
    return dval;

}


/*----------------------------------------------------------------------------*/
/**
 * @brief    Returns the requested element of a string list as a float
 *          
 *
 * @param  strlst                 The string list
 * @param  idx                    The element index 
 
*/
/*----------------------------------------------------------------------------*/

float mf_strlst_get_float (const char* strlst, int idx) {

    /* Get the list element as a double */
    double dval= mf_strlst_get_double (strlst,idx);

   /* Convert to a float */
   float fval = (float) dval;
  
   return fval;

}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Returns the requested element of a string list as an integer
 *           
 *
 * @param  strlst                 The string list
 * @param  idx                    The element index 
 
*/
/*----------------------------------------------------------------------------*/


int mf_strlst_get_int (const char* strlst, int idx) {

    /* Get the list element as a double */
    double dval= mf_strlst_get_double (strlst,idx);

   /* Convert to a float */
   int ival = (int) dval;
  
   return ival;

}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Returns the requested element of a string list as a boolean (1 or 0)
 *  
 *
 * @param  strlst                 The string list
 * @param  idx                    The element index 
 
*/
/*----------------------------------------------------------------------------*/

cpl_boolean mf_strlst_get_boolean (const char* strlst, int idx) {

    /* Get the list element as a double */
    double dval= mf_strlst_get_double (strlst,idx);

   /* Convert to an int */
   int ival = (int) dval;
  
   if (ival!=0) return CPL_TRUE;
  
   return CPL_FALSE;

}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Returns the requested element of a string list as a string trimmed
 *           of any leading and trailing whitespaces.
 * 
 *
 * @param  strlst                 The string list
 * @param  idx                    The element index 
 
*/
/*----------------------------------------------------------------------------*/

cpl_boolean mf_strlst_get_str (const char* strlst, int idx, char* ret_str) {

    int strsize=strlen(strlst);
    int nelems=mf_strlst_get_size(strlst);
    
    if (idx>nelems) return CPL_FALSE;
    
    /* Define the first and last deliminator position as -1 and string length*/
    int delim_pos[nelems+1];
    delim_pos[0]=-1;
    delim_pos[nelems]=strsize;
    
    /* Now locate the deliminator positions in the string itself*/
    int cnt=0;
    for (int i=0;i<strsize;i++) {
        if (strlst[i]!=',') continue;
        cnt=cnt+1;
        delim_pos[cnt]=i;
    }

    /* Define the indecies of interest */
    int idx1=delim_pos[idx-1]+1;
    int idx2=delim_pos[idx-0]-1;
    
    /*Alter first index so as to ignore leading whitespaces*/
    int new_idx1=idx1;
    for (int i=idx1;i<idx2; i++) {
        if (strlst[i]==' ') {new_idx1++;} else {break;}
    }
    idx1=new_idx1;

    /*Alter second index so as to ignore trailing whitespaces*/
    int new_idx2=idx2;
    for (int i=idx2;i>idx1; i--) {
        if (strlst[i]==' ') {new_idx2--;} else {break;}
    }
    idx2=new_idx2;

    /* Determine the character size of this element */
    int char_size=idx2-idx1+1;
    if (char_size<1)  char_size=1;
        
    /* Allocate a return string of appropriate size */
   /* char * ret_str = (char*) malloc((char_size+1)*sizeof(char));*/
   /* char * ret_str =  cpl_calloc((char_size+2),sizeof(char));*/
    
    /* Copy the element string to the return string one character at a time*/
    int j=0;
    for (int i=idx1;i<=idx2;i++) {
        ret_str[j]=strlst[i];
        j++;
    }
    
    /* Append the \0 char*/
    ret_str[char_size]='\0';
    
    /* Remove leading and trailing whaite spaces */
    /*ret_str=trim_space(ret_str); */
        
    return CPL_TRUE;

}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given fits data contains tables with a column of given name
 *
 * @param molecfit_fits* data         Loaded fits data.
 * @param char*          col_name     Name of column to look for
 *
 * @return             Error if column does not exist
 *
 *
 */
/* ---------------------------------------------------------------------------*/


cpl_error_code mf_config_chk_column(
    const molecfit_fits* fits_data,
    const char* col_name)
{
    if (fits_data==NULL) return CPL_ERROR_ILLEGAL_INPUT;
    if (col_name ==NULL) return CPL_ERROR_ILLEGAL_INPUT;
    
    
    if (mf_strlst_is_null(col_name)) {
        /* Column name is null. This is ok */
        return CPL_ERROR_NONE;
    } 


    /* Check that there are extensions */
    if (!fits_data->v_ext) {
        cpl_msg_error(cpl_func,"No extensions found in target frame with column name: %s!",col_name);        
        cpl_error_code err=CPL_ERROR_ILLEGAL_INPUT;
        cpl_error_set_message(cpl_func,err,"Invalid science frame!");
        return err;    
    }
    
    /* Iterate through each (non primary) extension*/

    cpl_boolean match_found_in_all_extensions = CPL_TRUE;
    for (cpl_size ext = 1; ext < fits_data->n_ext; ext++) {

        /* Skip if this extension does not have a header */
        if (!fits_data->v_ext[ext].header) continue;


        /* Find all TTYPE[n] components in the header */
        /* Note: this will be stored as a cpl_property element */
        const cpl_propertylist* pl = fits_data->v_ext[ext].header;
        cpl_size nprops=cpl_propertylist_get_size(pl);

        /* Iterate through all header property items looking for a name */
        /*   that starts "TTY"                                          */
        cpl_boolean match_found=CPL_FALSE;
        for (cpl_size i=1; i< nprops;i++) {

           const cpl_property *p = cpl_propertylist_get_const(pl,i);
           const char* name = cpl_property_get_name(p);
           if (name[0]=='T' &&
               name[1]=='T' &&
               name[2]=='Y' ) {
               
                 /* This is a TTYP[n} header keyword now get the string value */
                 const char* str_val = cpl_property_get_string(p);
               
                 /* Check if this matches the given column name to look for */
                 if (strcmp(str_val,col_name)==0) {
                     cpl_msg_info(cpl_func,"Looking for column %s in extension %lld ... Found.",col_name,ext);
                     match_found=CPL_TRUE;
                 }
            } /* end if name = "TTY*"*/

        } /* end for i*/
        
        if (!match_found) {
            cpl_msg_info(cpl_func,"Looking for column %s in extension %lld ... Not found.",col_name,ext);
            match_found_in_all_extensions = CPL_FALSE;
        }

     } /* end for ext */ 

   
    if (!match_found_in_all_extensions) {
        cpl_msg_error(cpl_func,"Could not find column %s in any extension of the data frame!",col_name);
        cpl_error_code err=CPL_ERROR_ILLEGAL_INPUT;
        cpl_error_set_message(cpl_func,err,"Could not find column %s as required!",col_name);
        return err;    
    }

    
    return CPL_ERROR_NONE;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given fits data contains the specific header keyword
 *
 * @param molecfit_fits* data         Loaded fits data.
 * @param char*          keyword      Name of keyword to look for
 *
 * @return             Error if keyword is not present.
 *
 *
 */
/* ---------------------------------------------------------------------------*/

cpl_error_code mf_config_chk_keyword(
    const molecfit_fits* fits_data,
    const char* keyword)
{
    if (fits_data==NULL) return CPL_ERROR_ILLEGAL_INPUT;
    if (keyword  ==NULL) return CPL_ERROR_ILLEGAL_INPUT;
    cpl_size keyword_len = strlen(keyword);
    if (keyword_len<1)   return CPL_ERROR_ILLEGAL_INPUT;

    if (keyword_len==4)  {
        if ( (keyword[0]=='N' || keyword[0]=='n') &&
             (keyword[1]=='O' || keyword[1]=='o') &&
             (keyword[2]=='N' || keyword[2]=='n') &&
             (keyword[3]=='E' || keyword[3]=='e') ) {
            return CPL_ERROR_NONE;
        }
    }
    
    /* Check that there are extensions */
    if (!fits_data->v_ext) {
        cpl_msg_info(cpl_func,"No extensions found in target frame!");        
        return CPL_ERROR_ILLEGAL_INPUT;    
    }
    
    /* Iterate through each (non primary) extension*/

    cpl_size ext = 0;

    /* Skip if this extension does not have a header */
    if (!fits_data->v_ext[ext].header)  {
        cpl_msg_info(cpl_func,"No primary header found in target frame");        
        return CPL_ERROR_ILLEGAL_INPUT;    
    }


    /* Note: this will be stored as a cpl_property element */
    const cpl_propertylist* pl = fits_data->v_ext[ext].header;
    cpl_size nprops=cpl_propertylist_get_size(pl);

    /* Iterate through all header property items looking for a name */
    /*   that ends with keyword, i.e. "*[keyword]                   */
    for (cpl_size i=1; i< nprops;i++) {

       const cpl_property *p = cpl_propertylist_get_const(pl,i);
       const char* name = cpl_property_get_name(p);
       cpl_size name_len = strlen(name);
       if (name_len!=keyword_len) continue;
       if (strcmp(keyword,name)==0) {
           cpl_msg_info(cpl_func, "Looking for header keyword %s\t....Found",keyword);
           return CPL_ERROR_NONE;
       }
     }
     
    cpl_msg_info(cpl_func, "Looking for header keyword %s\t....Not Found!",keyword);
    cpl_msg_info(cpl_func, "Unable to find keyword %s in target file header!",keyword);   
    return CPL_ERROR_ILLEGAL_INPUT;


}


/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given fits data contains the specific header keyword
 *
 * @param molecfit_fits* data         Loaded fits data.
 * @param char*          keyword      Name of keyword to look for
 *
 * @return             Error if keyword is not present.
 *
 *
 */
/* ---------------------------------------------------------------------------*/

cpl_error_code mf_config_check_keyword(
    const molecfit_fits* fits_data,
    const cpl_parameterlist* parlist,
    const char* keyword_par)
{
    if (fits_data    ==NULL) return CPL_ERROR_ILLEGAL_INPUT;
    if (keyword_par  ==NULL) return CPL_ERROR_ILLEGAL_INPUT;
    const cpl_parameter* par = cpl_parameterlist_find_const(parlist,keyword_par);
    const char* keyword = cpl_parameter_get_string(par);
    
    cpl_size keyword_len = strlen(keyword);
    if (keyword_len<1)   return CPL_ERROR_ILLEGAL_INPUT;

    if (keyword_len==4)  {
        if ( (keyword[0]=='N' || keyword[0]=='n') &&
             (keyword[1]=='O' || keyword[1]=='o') &&
             (keyword[2]=='N' || keyword[2]=='n') &&
             (keyword[3]=='E' || keyword[3]=='e') ) {
            return CPL_ERROR_NONE;
        }
    }
    
    
    if (keyword_len==3)  {
        if ( (keyword[0]=='A' || keyword[0]=='a') &&
             (keyword[1]=='N' || keyword[1]=='n') &&
             (keyword[2]=='Y' || keyword[2]=='y') ) {
            return CPL_ERROR_NONE;
        }
    }
    
    
    /* Check that there are extensions */
    if (!fits_data->v_ext) {
        cpl_msg_info(cpl_func,"No extensions found in target frame!");        
        return CPL_ERROR_ILLEGAL_INPUT;    
    }
    
    /* Iterate through each (non primary) extension*/

    cpl_size ext = 0;

    /* Skip if this extension does not have a header */
    if (!fits_data->v_ext[ext].header)  {
        cpl_msg_info(cpl_func,"No primary header found in target frame");        
        return CPL_ERROR_ILLEGAL_INPUT;    
    }


    /* Note: this will be stored as a cpl_property element */
    const cpl_propertylist* pl = fits_data->v_ext[ext].header;
    cpl_size nprops=cpl_propertylist_get_size(pl);

    /* Iterate through all header property items looking for a name */
    /*   that ends with keyword, i.e. "*[keyword]                   */
    for (cpl_size i=1; i< nprops;i++) {

       const cpl_property *p = cpl_propertylist_get_const(pl,i);
       const char* name = cpl_property_get_name(p);
       cpl_size name_len = strlen(name);
       if (name_len!=keyword_len) continue;
       if (strcmp(keyword,name)==0) {
           cpl_msg_info(cpl_func, "Looking for header keyword %s\t....Found",keyword);
           return CPL_ERROR_NONE;
       }
     }
     
    cpl_msg_info(cpl_func, "Looking for header keyword %s\t....Not Found!",keyword);
    return cpl_error_set_message(cpl_func,CPL_ERROR_ILLEGAL_INPUT,
                                 "Unable to find keyword %s in target file header!",keyword);


}



/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given ref atmos file lists all required molecules.
 *
 * @param char*      dirname          Path to Telliccorr data
 * @param char*      filename         Name of chosen Ref Atmos fits file
 * @param cpl_table* molecules        Table of molecules given as input parameters
 *
 * @return           Error if any input molecule is not listed in the Ref Atmos file
 *
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_error_code mf_config_chk_ref_atmos(
    const char* dirname,
    const char* filename,
    const cpl_table* molecules)
{

   /* Check input */
   if(!dirname)      return CPL_ERROR_ILLEGAL_INPUT;
   if(!filename)     return CPL_ERROR_ILLEGAL_INPUT;
   if(!molecules) {
       cpl_msg_info(cpl_func,"No molecules!");
       return CPL_ERROR_ILLEGAL_INPUT;
   }
   
   /* Get the actual Ref Atmos fits file */
   const char * subpath = "/profiles/mipas/";

   char * fullpathname = cpl_sprintf("%s%s%s", dirname,subpath,filename);
   cpl_msg_info(cpl_func,"Checking Reference Atmosphere file: %s", fullpathname);
   molecfit_fits* rat_data=NULL;
   
   /* Load up Ref Atmos fits file into table data */
   rat_data=molecfit_fits_load(fullpathname,CPL_FALSE);

   /* Check the Ref Atmos table data */
   if (!rat_data) {
       cpl_msg_info(cpl_func,"Unable to open rat data file %s!",fullpathname);
       cpl_free(fullpathname);
       return CPL_ERROR_ILLEGAL_INPUT;
   }

   cpl_free(fullpathname);
   /* Search for a bintable within this data*/
   cpl_array* column_names = NULL;
   cpl_size n_cols=0;
   for (cpl_size ext = 0; ext < rat_data->n_ext; ext++) {

      /* Skip if this extension does not have a bintable */
      if (!rat_data->v_ext[ext].table) continue;
      
      /* Get the table details */
      const cpl_table* table = rat_data->v_ext[ext].table;      
      n_cols = cpl_table_get_ncol(table);
      column_names=cpl_table_get_column_names(table);
      break;
   }

   /* Check that we have a valid bin table from the fits file */
   if (!column_names) {
       cpl_msg_info(cpl_func,"Unable to find Molecular data in %s",filename);
       return CPL_ERROR_ILLEGAL_INPUT; 
   }
   
   /* Iterate through the columns and list all molecules */
   /* (note this is all other than "HGT", "PRE" and "TEM")     */
   char **colname = cpl_array_get_data_string(column_names);
   cpl_size max_size = n_cols*10;
   char str_lst[max_size];
   str_lst[0]='\0';
   for (cpl_size col=0; col<n_cols; col++) {
       const char * name = colname[col];
       if (strcmp(name,"HGT")==0) continue;
       if (strcmp(name,"PRE")==0) continue;
       if (strcmp(name,"TEM")==0) continue;
       if (str_lst[0]!='\0') strcat(str_lst,",");
       strcat(str_lst,name);
   }   

   cpl_msg_info(cpl_func,"%s contains details on the following molecules:",filename);
   cpl_msg_info(cpl_func,"%s", str_lst);
   
   /** Now lets go through each molecule in the molecule table */
   cpl_size n_mols = cpl_table_get_ncol(molecules);
   n_mols=cpl_table_get_nrow (molecules);
   cpl_boolean all_fine=CPL_TRUE;
   for (cpl_size i=0; i<n_mols;i++) {
   
       /* Get the molecule name at this table index*/
       const char* mol_name = cpl_table_get_string (molecules,"LIST_MOLEC",i);
   
       /* Is this molecule in the file list of molecules */
       cpl_boolean isok = molecfit_config_str_isin_strlst(mol_name,str_lst);
       if (isok)    cpl_msg_info(cpl_func,"\tDoes list contain molecule %s ... Yes",mol_name);
       if (!isok)   cpl_msg_info(cpl_func,"\tDoes list contain molecule %s ... No!",mol_name);
       if (!isok)   all_fine=CPL_FALSE;
   }
   
   /* Free up the allocations */
   if (column_names)   cpl_free(column_names);
   if (rat_data)       cpl_free(rat_data);

   /* Report error if not fine */
   if (!all_fine) {
       cpl_msg_info(cpl_func,"Not all requested molecules are contained in the atmos ref file");
       return CPL_ERROR_ILLEGAL_INPUT;
   }

   return CPL_ERROR_NONE;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that the 3 molecule parameter strings are consistent
 *
 * @param char*      str1        List of molecule names
 * @param char*      str2        List of molecule fitting flags
 * @param char*      str3        List of molecule rel values.
 *
 * @return           Error if incompatible
 *
 *
 */
/* ---------------------------------------------------------------------------*/

cpl_error_code molecfit_config_molecule_strs_chk(const char* str1,const char* str2,const char* str3)

{
    /* Check that strings are valid*/ 
    if (!str1) {
        cpl_msg_info(cpl_func, "No external molecule table defined nor any valid molecule list parameters"); 
        return CPL_ERROR_ILLEGAL_INPUT;
    } 

    if (!str2) {
        cpl_msg_info(cpl_func, "No external molecule table defined nor any valid molecule list parameters"); 
        return CPL_ERROR_ILLEGAL_INPUT;
    } 

    if (!str3){
        cpl_msg_info(cpl_func, "No external molecule table defined nor any valid molecule list parameters"); 
        return CPL_ERROR_ILLEGAL_INPUT;
    } 

    if (mf_strlst_is_null(str1)) {
        cpl_msg_info(cpl_func, "No external molecule table defined nor any valid molecule list parameters"); 
        return CPL_ERROR_ILLEGAL_INPUT;
    } 

    if (mf_strlst_is_null(str2)) {
        cpl_msg_info(cpl_func, "No external molecule table defined nor any valid molecule list parameters"); 
        return CPL_ERROR_ILLEGAL_INPUT;
    } 

    if (mf_strlst_is_null(str3)) {
        cpl_msg_info(cpl_func, "No external molecule table defined nor any valid molecule list parameters"); 
        return CPL_ERROR_ILLEGAL_INPUT;
    } 

    /* Check that string lists have same sizes*/ 
    int n1 = mf_strlst_get_size(str1);
    int n2 = mf_strlst_get_size(str2);
    int n3 = mf_strlst_get_size(str3);

    if (n1!=n2 || n1!=n3 || n2!=n3) {
        cpl_msg_info(cpl_func, "Molecule parameter lists have incompatible sizes"); 
        cpl_msg_info(cpl_func, "\t%s",str1); 
        cpl_msg_info(cpl_func, "\t%s",str2); 
        cpl_msg_info(cpl_func, "\t%s",str3);
        return CPL_ERROR_ILLEGAL_INPUT;    
    }


    return CPL_ERROR_NONE;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given gdas_prof is a valid values.
 *
 * @param char*      gdas_prof        Supplied value for gdas profile
 *
 * @return           Error if specified file cannot be found.
 *
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_error_code mf_config_chk_gdas_prof(const char* gdas_prof)
{
    /* Check that gdas_prof is not "auto" or "none" */
    if (strcmp(gdas_prof,"auto")==0) return CPL_ERROR_NONE;
    if (strcmp(gdas_prof,"AUTO")==0) return CPL_ERROR_NONE;
    if (strcmp(gdas_prof,"none")==0) return CPL_ERROR_NONE;
    if (strcmp(gdas_prof,"NONE")==0) return CPL_ERROR_NONE;
    if (strcmp(gdas_prof,"null")==0) return CPL_ERROR_NONE;
    if (strcmp(gdas_prof,"NULL")==0) return CPL_ERROR_NONE;

    /* Assume that this is a given filename and check that it exists*/
    cpl_error_code file_err = access(gdas_prof, F_OK);

    if (!file_err) cpl_msg_info(cpl_func,"GDAS Profile file %s Exists .... Yes",gdas_prof);
    if ( file_err) cpl_msg_info(cpl_func,"GDAS Profile file %s Exists .... No!",gdas_prof);
    if ( file_err) cpl_msg_info(cpl_func,"Invalid GDAS Profile. Cannot find file %s",gdas_prof);

    return file_err;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given lnfl line data base exists.
 *
 * @param char*      dirname          Path to Telliccorr data
 * @param char*      db               Name of chosen database file
 *
 * @return           Error if any input molecule is not listed in the Ref Atmos file
 *
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_error_code mf_config_chk_lnfl_line_db(
    const char* dirname, 
    const char* db)
{
    /* Check input */
    if(!dirname)      return CPL_ERROR_ILLEGAL_INPUT;
    if(!db     )      return CPL_ERROR_ILLEGAL_INPUT;

    /* Get the actual lnfl_line_db file pathname */
    const char * subpath = "/hitran/";

    char * fullpathname = cpl_sprintf("%s%s%s", dirname,subpath,db);
    cpl_msg_info(cpl_func,"Checking LNFL Line DataBase file: %s", fullpathname);

    /* Assume that this is a given filename and check that it exists*/
    cpl_error_code file_err = access(fullpathname, F_OK);

    if (!file_err) cpl_msg_info (cpl_func,"LNFL database  %s Exists .... Yes",db);
    if ( file_err) cpl_msg_info (cpl_func,"LNFL database  %s Exists .... No!",db);
    if ( file_err) cpl_msg_error(cpl_func,"Invalid LNFL database %s",db);

    cpl_free(fullpathname);
    return file_err;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given wlc_n and wlc_const are compatible strings.
 *
 * @param char*      wlc_n_str          wlc poly order string list
 * @param char*      wlc_c_str          wlc const value string list
 * @param char*      nchips             No of actual chips
 *
 * @return           Error if any input molecule is not listed in the Ref Atmos file
 *
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_error_code  molecfit_config_mappingchk(
    const cpl_frameset* frameset,
    const cpl_parameterlist* parlist,
    const char*  library,
    const char*  lib_map,
    const char*  ext_column,
    const char*  map_parname,
    const char*  target_tag)
 {
    cpl_msg_info(cpl_func,"Validating any declared mappings to %s ...",library);

    cpl_boolean chk_numbers_match=CPL_TRUE;
    if (!strcmp(target_tag,"NULL")) chk_numbers_match=CPL_FALSE;

    /* Check the input parameters */
    if (!frameset) {
        cpl_error_code err =CPL_ERROR_ILLEGAL_INPUT;
        cpl_msg_error(cpl_func,"Invalid frameset!");
        cpl_error_set_message(cpl_func,err,"Error in handling frameset!");
        return err;
    }
  
    if (!parlist) {
        cpl_error_code err =CPL_ERROR_ILLEGAL_INPUT;
        cpl_msg_error(cpl_func,"Invalid parameter list!");
        cpl_error_set_message(cpl_func,err,"Error in handling parameter list!");
        return err;
    }
    
    cpl_boolean chip_extensions=CPL_FALSE;
    const cpl_parameter* p = cpl_parameterlist_find_const(parlist, MOLECFIT_PARAMETER_CHIP_EXTENSIONS);
    if (parlist)  chip_extensions = cpl_parameter_get_bool(p);
    if (chip_extensions && strcmp(target_tag,"SCIENCE")==0) {
        cpl_msg_info(cpl_func,"\t Using chip extensions with SCIENCE frame so skipping SCIENCE related mapping checks");
        return CPL_ERROR_NONE;
    }


    /* FIRST CHECK THAT THERE IS A LIBRARY FILE WITH EXTENSIONS */
    const cpl_frame *frame    = cpl_frameset_find_const (frameset, library);       
    if (!frame   ) {
        cpl_msg_info(cpl_func,"\tThere is no specified frame with tag %s so nothing to check",library);
        return CPL_ERROR_NONE;
    }
    
    /* Get the no of extensions in this library frame */    
    cpl_size n_ext = cpl_frame_get_nextensions(frame);
    const char* str;
    str = (n_ext==1) ? "extension" :  "extensions";
    cpl_msg_info(cpl_func,"\tFound frame %s with %lld %s",library,n_ext,str);


    /* NOW CHECK FOR ANY MAPPINGS TO THE LIBRARY EXTENSIONS */

    cpl_msg_info(cpl_func,"\tChecking that the %s parameter declares a valid  mapping to %s:",map_parname,library); 
    if (chk_numbers_match) cpl_msg_info(cpl_func,"\tChecking that the numbers match.");

    const cpl_parameter *par;

    
    

    
    /* Now check for a MAPPING KERNEL declaration in the MODEL_MAPPING_KERNEL parameter of the.rc file */
    /*parname = MOLECFIT_MODEL_MAPPING_KERNE*/;
    par=cpl_parameterlist_find_const(parlist,map_parname);    
    if (!par) {
        cpl_error_code err =CPL_ERROR_ILLEGAL_INPUT;
        cpl_msg_error(cpl_func,"Cannot find parameter %s",map_parname);
        cpl_error_set_message(cpl_func,err,"Error in parameter %s appears to be missing!",map_parname);
        return err;
    }

    const char * string = cpl_parameter_get_string(cpl_parameterlist_find_const(parlist,map_parname));
    
    /* We have a string value for the MAPPING  now check it */    
    cpl_error_code mapping_err=CPL_ERROR_NONE;
    cpl_size n_mappings;
    cpl_size n_non_primary_mappings;


    /* Check if this string is "NULL" in which case look up for a mapping file or a string list of values*/
    if (mf_strlst_is_null(string)) {
    
        /* Parameter value is NULL so look for a mapping file */

        const cpl_frame *map_frame = cpl_frameset_find_const(frameset, lib_map);       
        const char *filename = cpl_frame_get_filename(map_frame);
        cpl_table* map_table=NULL;
        if (filename) map_table = cpl_table_load(filename, 1, 0);
        
        if (!map_table) {
        
            cpl_msg_info(cpl_func,"\t%s String is null. Looking for a mapping frame %s  .... Not Found!", map_parname,lib_map);
            cpl_msg_error(cpl_func,"\tUnable to find mapping table in valid %s frame!",lib_map);
            return CPL_ERROR_ILLEGAL_INPUT;

        } 
        cpl_msg_info(cpl_func,"\t%s String is null. Looking for a mapping frame %s  .... Found", map_parname,lib_map);

        /*cpl_table_dump(map_table, 0, cpl_table_get_nrow(map_table), NULL);                */
        cpl_msg_info(cpl_func,"\t\tChecking main table values of %s are within bounds.",lib_map);

        /* Extract the extension values from the column of the form: _EXT */
        /* and check that they are within bounds */
        cpl_size n = cpl_table_get_nrow(map_table);        
        cpl_size n_non_zero=0;
        for (cpl_size i=0;i<n;i++) {
           /*int ival = cpl_table_get_int (mapping_kernel_tst,"KERNEL_LIBRARY_EXT",i,NULL);*/
           long long llval=cpl_table_get_long_long (map_table,ext_column,i,NULL);
           if (llval<0) {
               cpl_msg_error(cpl_func,"Row %lld of the MAPPING TABLE has -ve extension value %lld",i+1,llval);
               mapping_err=CPL_ERROR_ILLEGAL_INPUT;
           }
           if (llval>n_ext) {
               cpl_msg_error(cpl_func,"Row %lld of the MAPPING TABLE has extension value %lld which is > than number of extensions in Library (%lld)",i+1,llval,n_ext);
               mapping_err=CPL_ERROR_ILLEGAL_INPUT;
           }
           if (llval>0 && llval <= n_ext) n_non_zero++; 
           /*cpl_msg_info(cpl_func,"Row %lld = %lld", i, llval);           */
        }
        if (!mapping_err) cpl_msg_info(cpl_func,"\t\tAll mappings are to valid extension numbers.");

        if (map_table) cpl_free( map_table) ;
        
        n_mappings=n;
        n_non_primary_mappings=n_non_zero;

    } else {
    
        /* Parameter value is a string list so check its integer elements are within bounds */

        int nelems = mf_strlst_get_size(string);
        int n_non_zero=0;
        for (int i = 1; i<=nelems ; i++ ) {
            int ival = mf_strlst_get_int(string,i);
            if (ival<0) {
                cpl_msg_error(cpl_func,"Item %d of the MAPPING string list has -ve extension value %d",i+1,ival);
                mapping_err=CPL_ERROR_ILLEGAL_INPUT;
            }
            if (ival>n_ext) {
                cpl_msg_error(cpl_func,"Item %d of the MAPPING string list has extension value %d which is > than number of extensions in Library (%lld)",i+1,ival,n_ext);
                mapping_err=CPL_ERROR_ILLEGAL_INPUT;
            }
            if (ival>0 && ival <= n_ext) n_non_zero++;
        }
        
        n_mappings=nelems;
        n_non_primary_mappings=n_non_zero;           
    }
        
   
   /* Report if there is a mapping error */ 
   if (mapping_err) {
       cpl_error_set_message(cpl_func,mapping_err,"Invalid Mapping Definition %s",lib_map); 
   }
   
   
   /* If requested check that the numbers match */ 
   if (!mapping_err && chk_numbers_match) {
       cpl_msg_info(cpl_func,"\t\tChecking that there are the same no of non primary mappings as there are extensions in %s",target_tag);
       const cpl_frame *target_frame    = cpl_frameset_find_const(frameset, target_tag);       

       if (!target_frame   ) {
           cpl_msg_info(cpl_func,"\t\tThere is no specified frame with tag %s so nothing to check",target_tag);
           return CPL_ERROR_NONE;
       }
        /* Get the no of extensions in this library frame */    
       cpl_size n_ext_target = cpl_frame_get_nextensions(target_frame);
       if (n_ext==n_non_primary_mappings) {
           cpl_msg_info(cpl_func,"\t\tFound target_frame %s with %lld extensions.   ... Numbers match.",target_tag,n_ext_target);
       } else {    
           cpl_msg_info(cpl_func,"\t\tFound target_frame %s with %lld extensions.   ... Numbers do not match!",target_tag,n_ext_target);
           cpl_msg_info(cpl_func,"\t\t Number of mappings found=%lld. Number of non primary mappings found=%lld",n_mappings,n_non_primary_mappings);
           cpl_msg_error(cpl_func,"\t\t\t Problems with mapping!");
           mapping_err=CPL_ERROR_ILLEGAL_INPUT;
           cpl_error_set_message(cpl_func,mapping_err,"Mapping mismatch for %s to %s",map_parname,target_tag);
              
       }
   }
   
    return  mapping_err;
 }
 
/* ---------------------------------------------------------------------------*/
/**
 * @brief Check that given wlc_n and wlc_const are compatible strings.
 *
 * @param char*      wlc_n_str          wlc poly order string list
 * @param char*      wlc_c_str          wlc const value string list
 * @param char*      nchips             No of actual chips
 *
 * @return           Error if any input molecule is not listed in the Ref Atmos file
 *
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_error_code mf_config_chk_wlc_strs(
    const char* wlc_n_str, 
    const char* wlc_c_str,
    int nchips)
    
{
   if (!wlc_n_str) cpl_msg_info(cpl_func,"Bad wlc_nstr string. N chips=%d",nchips);
   if (!wlc_c_str) cpl_msg_info(cpl_func,"Bad wlc_cstr string. N chips=%d",nchips);

   return CPL_ERROR_NONE;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Given a frame for SCIENCE data, check if the data format is a table or not
 *
 * @param cpl_frame *input_frame The science data frame being investigated
 *
 * @return           True if table type false otherwise
 *
 *
 */
/* ---------------------------------------------------------------------------*/

cpl_boolean  molecfit_config_chk_science_is_table(cpl_frame *input_frame) {

    const char * filename = cpl_frame_get_filename (input_frame);
    cpl_msg_info(cpl_func,"Checking data structure type of the SCIENCE file %s",filename);
    molecfit_fits_file_format format = molecfit_fits_get_file_format(filename,CPL_FALSE);
    switch(format) {

        case molecfit_fits_file_table :
            cpl_msg_info(cpl_func,"\t Data structure is in a molecfits table format");
            break;
        case molecfit_fits_file_image_1D :
            cpl_msg_info(cpl_func,"\t Data structure is a 1D image");
            break;
        case molecfit_fits_file_image_2D :
            cpl_msg_info(cpl_func,"\t Data structure is a 2D image");
            break;
        case molecfit_fits_file_image_3D :
            cpl_msg_info(cpl_func,"\t Data structure is a 3D image");
	        break;
        default :
            cpl_msg_info(cpl_func,"\t Data structure is a Something else!");
            break;
    }
    
    if (format!=molecfit_fits_file_table) return CPL_FALSE;
    
    return CPL_TRUE;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Split a string into pieces at a given delimiter.
 *
 * @param str         The string to split.
 * @param delimiter   String specifying the locations where to split.
 * @param size        (Output) If non NULL, the number of found tokens is returned.
 *
 * @return The function returns a newly allocated, @c NULL terminated array of strings, or @c NULL in case of an error.
 *
 * The function breaks up the string @em string into pieces at the places indicated by @em delimiter.
 * The delimiter string @em delimiter never shows up in any of the resulting strings.
 *
 * As a special case, the result of splitting the empty string "" is an empty vector, not a vector containing a single string.
 *
 * The created result vector can be deallocated using @b mf_string_strfreev().
 *
 */
/* ---------------------------------------------------------------------------*/
static char ** molecfit_config_str_split(
    const char               *str,
    const char               *delimiter,
    cpl_size                 *size)
{
  /* Check inputs */
  cpl_error_ensure(str && delimiter && size, CPL_ERROR_NULL_INPUT,
                   return NULL, "Not all input data is provided!");
  cpl_error_ensure(*delimiter != '\0', CPL_ERROR_ILLEGAL_INPUT,
                   return NULL, "delimiter is \"\\0\"!");

  /* Get the number of tokens to allocate */
  unsigned int sz        = strlen(delimiter);
  const char   *save_str = str;
  char         *s        = strstr(save_str, delimiter);
  cpl_size     nr        = 1;

  while (s && (strlen(s) > sz)) {
      save_str = s + sz;
      s         = strstr(save_str, delimiter);
      nr++;
  }

  /* Allocate array */
  char **sarray = cpl_calloc(nr + 1, sizeof(char*));
  *size = nr;

  /* Copy tokens into array */
  cpl_size i = 0;
  save_str   = str;
  s          = strstr(save_str, delimiter);

  if (!s || !(strlen(s) > sz)) {

      /* Only one token, copy it */
      sarray[i] = cpl_sprintf("%s", str);

  } else {

      while (s && (strlen(s) >= sz)) {

          cpl_size length = s - save_str;

          /* Initialize token */
          int n = length + 1;
          char *token = cpl_calloc(n, sizeof(char));
          strncpy(token, save_str, length);
          token[length] = '\0';

          /* Put it into array */
          sarray[i] = token;
          i++;

          /* Prepare next token */
          save_str = s + sz;
          if (strlen(save_str) == 0) s = NULL;
          else                       s = strstr(save_str, delimiter);
      }

      /* Copy last token if there is one and if it isn't the delimiter */
      if (save_str && strlen(save_str) > 0 && strcmp(save_str, delimiter) != 0) {
        sarray[i] = cpl_sprintf("%s", save_str);
      } else {
        sarray[i] = NULL;
      }
  }

  /* Initialize last member to NULL */
  sarray[nr] = NULL;

  return sarray;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Deallocate a string array.
 *
 * @param strarr  String array to deallocate
 *
 * The function deallocates the array of strings @em strarr and any string it possibly contains.
 *
 */
/* ---------------------------------------------------------------------------*/
static void molecfit_config_str_array_delete(
    char                     **strarr)
{
  if (strarr) {
      cpl_size i = 0;
      while (strarr[i] != NULL) {
          cpl_free(strarr[i]);
          i++;
      }
      cpl_free(strarr);
  }
}

/** @cond PRIVATE */

/** @endcond */


/**@}*/


